# -*- encoding: utf-8 -*-
#

"""
This is Wombat, an operations and maintenance tool for Erlang systems
© 2015, Erlang Solutions Ltd
"""

from __future__ import absolute_import

from wombat.alarm       import Alarm
from wombat.log         import Log
from wombat.node_family import NodeFamily
from wombat.node        import Node
from wombat.utils       import timestr
from wombat.wombat_rest import wombat_opener, rest_call, quote_path, quote_args

class WombatException(Exception):
    pass

class Wombat(object):
    """This class represent a WombatOAM instance"""

    def __repr__(self):
        return 'Wombat({0.host!r}, {0.port!r})'.format(self)

    def __str__(self):
        return 'Wombat at {0.host!s}:{0.port!s}'.format(self)

    def __init__(self, host = 'localhost', port = 8080,
                 user='admin', password='admin', https = False):
        """Initialize with the address of the REST API"""
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        if https: self.proto = 'https'
        else: self.proto = 'http'

        self.opener = wombat_opener(self.__url(''), user, password)
        [header, rest] = self.get('help/about')['text'].split('\n\n\n', 1)
        self.header = header
        self.dashboard_url = '{0.proto}://{0.host}:{0.port}/'.format(self)

    def get_node_families(self, value = None, field = 'name'):
        """Return the list of NodeFamily objects present in Wombat.

        This funtion takes optional arguments for filtering the
        results for the value of a field (by default: the name
        field). For example get_node_families('Erlang/OTP 17') will
        return only the node families named 'Erlang/OTP 17', and
        get_node_families(true, field = 'deleted') will only return
        deleted node families.

        """
        return [NodeFamily(self, node_family['id'], node_family)
                for node_family in self.get('topo/node-family') \
                if value is None or node_family[field] == value]

    def get_node_family(self, value, field = 'name'):
        """Return a NodeFamily object by its name.

        If the name is not unique or does not exist, this function
        raises a WombatException.

        This function can also search by other fields, not just name
        if the optional field argument is given.

        """

        families = self.get_node_families(value, field)
        if len(families) == 1:
            return families[0]

        raise WombatException(
            'Looking for node family with {!s} {} ' \
            'returned {} results'.format(field, value, len(families)))

    def get_nodes(self, value = None, field = 'name'):
        """Return the list of Node objects present in Wombat.

        This funtion takes optional arguments for filtering the
        results for the value of a field (by default: the name
        field). For example get_nodes('riak@10.0.0.1') will return
        only the nodes named 'riak@10.0.0.1', and get_nodes('UP',
        field = 'state') will only return nodes currently connected to
        Wombat.

        """
        return [Node(self, node['id'], node)
                for node in self.get('topo/node') \
                if value is None or node[field] == value]

    def get_node(self, value, field = 'name'):
        """Return a Node object by its name.

        If the name is not unique or does not exist, this function
        raises a WombatException.

        This function can also search by other fields, not just name
        if the optional field argument is given.

        """

        nodes = self.get_nodes(value, field)
        if len(nodes) == 1:
            return nodes[0]

        raise WombatException(
            'Looking for node with {!s} {} ' \
            'returned {} results'.format(field, value, len(nodes)))

    def add_node(self, node_name, cookie, discover = False):
        """Add a new managed node to Wombat.

        By default only the provided node is added, but if discover is
        set to True, Wombat will also start managing the entire
        cluster the node belongs to.

        Without discovery a single Node object is returned, otherwise
        a list of Node objects is provided. In case of failure an
        exception is raised.

        """

        data = {'name':   node_name,
                'cookie': cookie}
        action = 'add-node-and-disc' if discover else 'add-node'
        code, body = self.post('topo/action/' + action, data)

        if code != 200:
            raise WombatException(
                'Failed to add node {} with cookie {} (HTTP error {}): {!s}'
                    .format(node_name, cookie, code, body))

        if discover:
            return [Node(self, node['id'], node) for node in body]
        else:
            node = body[0]
            return Node(self, node['id'], node)

    def delete_node(self, node):
        """Remove the node from being managed in Wombat."""
        self.delete('topo' + self.__get_scope(node, None))

    def delete_node_family(self, node_family):
        """Remove the node family from Wombat.

        This function fails if there are any nodes left in the node
        family.

        """
        self.delete('topo' + self.__get_scope(None, node_family))

    def analyse_reachability(self, node_name):
        """Analyse the possiblity of reaching the node via Distributed
        Erlang.

        This check is most useful to perform after an attempt to add
        the node to Wombat failed.

        """

        code, body = self.post('topo/action/analyse-reachability',
                               {'name': node_name})
        return body

    def get_metric_info(self, node = None, node_family = None):
        """Get metadata for all known metrics on a node or node family."""

        # Calculate scope
        scope = self.__get_scope(node, node_family)
        if not scope:
            raise ValueError('Invalid scope')

        return self.get('metric/metric-info' + scope)['groups']

    def get_gauge_info(self, node):
        """Get metadata for gauges defined on a node."""

        # Calculate scope
        scope = self.__get_scope(node, None)

        return self.get('metric/gauge-info' + scope)['gauges']

    def get_metric(self, metric_id, metric_type, from_time, to_time,
                   series = 1, node = None, node_family = None):
        """Fetch metric samples for a metric in the given time interval.

        Series only apply for counters and gauges. Also, these are the
        only metric types supported for node families.

        When metrics are queried for a node, a list of samples is
        returned. For a node family however the return value is a
        dictionary mapping node ids to list of samples.

        """

        # Calculate scope
        scope = self.__get_scope(node, node_family)
        if not scope:
            raise ValueError('Invalid scope')

        # The actual URL depends a lot on the metric type
        if metric_type == 'counter' or metric_type == 'gauge':
            url = 'metric/history{0}/{1}/{3}?{4}'
        else:
            url = 'metric/{2}{0}/{1}?{4}'
        url = url.format(scope,
                         quote_path(metric_id),
                         quote_path(metric_type),
                         quote_path(str(series)),
                         quote_args({'from': timestr(from_time),
                                     'to':   timestr(to_time)}))

        response = self.get(url)

        # Node and family level responses differ a bit.
        if node_family:
            result = {}
            for node in response:
                result[node['nodeId']] = node['data']
            return result
        else:
            return response['data']

    def get_node_metric(self, node, metric_type, metric,
                        to_time, from_time):
        """Fetch metric samples for a metric in the given time interval.

        Series can apply all kinds of metric, but only for nodes.
        """

        series = ''
        if metric_type == 'counter' or metric_type == 'gauge':
            series = '/1'
            metric_type = 'history'

        url = 'metric/{0}/node/{1}/{2}{3}?{4}'.format(
                quote_path(metric_type),
                quote_path(node),
                quote_path(metric),
                series,
                quote_args({'from': from_time, 'to': to_time}))
        return self.get(url)['data']

    def enum_logs(self, node = None, node_family = None, search = None,
                  originator = None, case_sensitive = False):
        """Return a generator for enumerating logs in Wombat.

        When no arguments are provided, all notifications are
        considered. But it is possible to limit the scope to a single
        node, node family or search filter (case sensitive or not).

        It is also possible to enumerate only logs coming from a given
        originator as well.

        """

        # Fixme: it would be nice to allow enumerating notifications
        # starting from a cursor position too - but returning a cursor
        # is difficult: for the first page of logs we have no cursors
        # at all, and starting from the second page we have cursors
        # for the beginning of the page only. Possible solutions could
        # be:
        #
        # - yield a cursor object at the end of every page: this would
        #   complicate iterating over logs way too complex
        #
        # - add a cursor field to the log that holds both Wombat's
        #   original cursor and an offset within the page: this would
        #   work, except for the log entries on the first page

        # Calculate scope
        scope = self.__get_scope(node, node_family)

        # Calculate arguments from the search and originator filters
        args = {}
        if search:
            # Search acts as a scope and mutually exclusive with node
            # and node_family!
            if scope:
                raise ValueError('Ambiguous scope')

            scope = '/search'
            args['q'] = search
            if case_sensitive:
                args['cs'] = True
        if originator:
            args['originator'] = originator

        # Enumerate notifications in a loop
        url = 'logs' + scope
        if not scope: 
            scope = '/all'
        while True:
            # Add the arguments to the REST url
            if args:
                url += '?' + quote_args(args)

            # Query & enumerate the next page of logs
            data = self.get(url)
            for log in data['logs']:
                yield Log(self, log)

            cursor = data.get('cursor')
            if not cursor:
                # End of notification list
                return

            # Prepare next REST url
            url = 'logs/previous' + scope
            args['c'] = cursor

    def get_log_originators(self, node = None, node_family = None):
        """Return the list of all log originators available for filtering the
        notifications.

        When no arguments are provided this function works on a global
        scope, but can be limited to list log originators on a
        specific node or node family too.

        """

        scope = self.__get_scope(node, node_family)
        return self.get('logs/originator' + scope)['originators']

    def get_log_levels(self, node):
        """Return a dictionary mapping handlers to log levels on the node."""

        url = 'logs/level/' + \
              quote_path(node.id if isinstance(node, Node) else node)

        levels = {}
        for item in self.get(url)['logLevels']:
            levels[item['handler']] = item['logLevel']

        return levels

    def set_log_level(self, node, handler, level):
        """Update the log level for the specified handler on a node."""

        url = 'logs/level/' + \
              quote_path(node.id if isinstance(node, Node) else node)
        data = {'handler': handler, 'level': level}

        code, body = self.post(url, data)

        if not code == 200:
            raise WombatException(
                'Failed to set log level to {} for {} of node {} in Wombat ' \
                '(HTTP error {}): {!s}'.format(level, handler, node,
                                               code, body))

    def enum_alarms(self, node = None, node_family = None, source = None,
                    following = None):
        """Return a generator for enumerating alarms in Wombat.

        When no arguments are provided, all alarms are considered. But
        it is possible to limit the scope to a single node or node
        family, or to filter for a specific alarm source.

        It is also possible to enumerate only alarms following a given
        event.

        """

        # Calculate scope
        scope = self.__get_scope(node, node_family)

        # Ensure following is an event id and not an Alarm object
        if isinstance(following, Alarm):
            following = alarm.eventId

        # Calculate arguments from the source filter
        if source is None:
            args = {}
        else:
            args = {'source': source}

        # Enumerate alarms in a loop
        while True:
            # Build the REST url
            url = 'alarm'
            if following is None:
                url += scope
            else:
                url += '/previous' + (scope if scope else '/all')
                args['eid'] = following

            if args:
                url += '?' + quote_args(args)

            # Query & enumerate the next page of alarms
            alarms = self.get(url)['alarms']
            for alarm in alarms:
                yield Alarm(self, alarm)

            if len(alarms) == 0:
                # End of alarm list
                return

            following = alarms[-1]['eventId']

    def get_alarm_sources(self, node = None, node_family = None):
        """Return the list of all alarm sources available for
        filtering the alarms.

        When no arguments are provided this function works on a global
        scope, but can be limited to list alarm sources on a specific
        node or node family too.
        """

        scope = self.__get_scope(node, node_family)
        return self.get('alarm/source' + scope)['sources']

    def acknowledge_alarm(self, alarm, alarm_src = None, user = None):
        """Acknowledge an alarm.

        The alarm to acknowledge can be provided as an Alarm object or
        as an alarm id. In the second case the raw source field of the
        alarm is needed as well.

        The user defaults to the user name used for authenticating
        towards Wombat but can be customised.

        """
        self.__update_alarm(alarm, alarm_src, user, 'acknowledge')

    def unacknowledge_alarm(self, alarm, alarm_src = None, user = None):
        """Unacknowledge an alarm.

        The alarm to unacknowledge can be provided as an Alarm object
        or as an alarm id. In the second case the raw source field of
        the alarm is needed as well.

        The user defaults to the user name used for authenticating
        towards Wombat but can be customised.

        """
        self.__update_alarm(alarm, alarm_src, user, 'unacknowledge')

    def clear_alarm(self, alarm, alarm_src = None, user = None):
        """Clear an alarm.

        The alarm to clear can be provided as an Alarm object or as an
        alarm id. In the second case the raw source field of the alarm
        is needed as well.

        The user defaults to the user name used for authenticating
        towards Wombat but can be customised.

        """
        self.__update_alarm(alarm, alarm_src, user, 'clear')

    def comment_alarm(self, message, alarm, alarm_src = None, user = None):
        """Comment on an alarm.

        The alarm to comment can be provided as an Alarm object or as
        an alarm id. In the second case the raw source field of the
        alarm is needed as well.

        The user defaults to the user name used for authenticating
        towards Wombat but can be customised.

        """
        self.__update_alarm(alarm, alarm_src, user, 'comment',
                            {'comment': message})

    def __update_alarm(self, alarm, alarm_src, user, action, data = {}):
        """Update an alarm."""

        # Calculate standard request data
        data['userId'] = user if user else self.user
        if isinstance(alarm, Alarm):
            data['alarmId']  = alarm.alarmId
            data['alarmSrc'] = alarm.rawSrc
        else:
            if alarm_src is None:
                raise ValueError('An alarm_src is required')

            data['alarmId']  = alarm
            data['alarmSrc'] = alarm_src

        # Send request
        code, body = self.post('alarm/' + action, data)
        if code != 200:
            raise WombatException(
                'Failed to {} alarm on Wombat (HTTP error {}): {!s}'
                    .format(action, code, body))

    def get_plugins(self, node):
        """Get the state of plugins on a managed node.

        The result is a dictionary mapping plugin names to plugin
        data.

        """
        
        scope = self.__get_scope(node, None)

        plugins = {}
        for plugin in self.get('plugins' + scope + '/plugin')['plugins']:
            plugins[plugin.pop('plugin_name')] = plugin

        return plugins

    def set_plugin_state(self, node, plugin, state):
        """Set the state of a plugin or all plugins on a managed node.

        The state should be 'on' or 'off'. If plugin is a string, only
        that plugin's state will be changed. But it can also be None,
        in which case all plugins are affected.

        """

        scope = self.__get_scope(node, None)

        url = 'plugins' + scope + '/plugin'
        if plugin:
            url += '/' + quote_path(plugin)
        data = {'plugin_state': state}

        code, body = self.put(url, data)
        if code != 200:
            raise WombatException(
                'Failed to set plugin state to {} for {} on Wombat ' \
                '(HTTP error {}): {!s}'
                    .format(state, plugin if plugin else 'all plugins',
                            code, body))

    @staticmethod
    def __get_scope(node, node_family):

        """Return an URL's scope part based on whether a node or a
        node family was specified.

        It is an error to specify both a node and a node family at the
        same time, but it is OK to leave both arguments None, which
        would mean a global scope.

        The node and node family may be specified either as an id
        (string) or as a Node or NodeFamily instance respectively.

        """
        if not node is None:
            if not node_family is None:
                raise ValueError('Ambiguous scope')

            return '/node/{}'.format(
                quote_path(node.id if isinstance(node, Node) else node))
        elif not node_family is None:
            return '/node-family/{}'.format(
                quote_path(node_family.id if
                           isinstance(node_family, NodeFamily) else
                           node_family))
        else:
            return ''

    def __url(self, path):
        """Format a path into a REST url"""
        return '{0.proto}://{0.host}:{0.port}/api/{1}'.format(self, path)

    def get(self, path, format = 'python'):
        """Do an HTTP GET towards Wombat.

        Returns the body of the response in the requested format.

        """
        url = self.__url(path)

        try:
            code, body = rest_call(url, self.user, self.password, format = format)
        except Exception as e:
            raise WombatException(
                'Error getting data from Wombat at {}. Reason: {!s}'
                    .format(url, e))

        if not code == 200:
            raise WombatException(
                'Error getting data from Wombat at {}. Reason: ' \
                'HTTP status code {}'.format(url, code))

        return body

    def delete(self, path):
        """Do an HTTP DELETE towards Wombat.

        Returns the HTTP status code.

        """
        url = self.__url(path)

        try:
            code, body = rest_call(url, self.user,self.password, method = 'DELETE',
                                   format = 'json')
            return code
        except Exception as e:
            raise WombatException(
                'Error deleting resource from Wombat at {}. Reason: {!s}'
                    .format(url, e))

    def post(self, path, data, format = 'python'):
        """Do an HTTP POST towards Wombat.

        Returns the HTTP status code and the body of the response in
        the requested format.

        """
        url = self.__url(path)

        try:
            return rest_call(url, self.user, self.password, data, format = format)
        except Exception as e:
            raise WombatException(
                'Error posting data {} to Wombat at {}. Reason: {!s}'
                    .format(data, url, e))

    def put(self, path, data, format = 'python'):
        """Do an HTTP PUT towards Wombat.

        Returns the HTTP status code and the body of the response in
        the requested format.

        """
        url = self.__url(path)

        try:
            return rest_call(url, self.user, self.password, data, method = 'PUT',
                             format = format)
        except Exception as e:
            raise WombatException(
                'Error putting data {} to Wombat at {}. Reason: {!s}'
                    .format(data, url, e))
