#! /usr/bin/env escript
%% -*- erlang -*-
%%! -s inets start -name node-restructure-${WOMBAT_NODENAME} -setcookie ${WOMBAT_COOKIE} -hidden

-define(WOMBAT_NODE, '${WOMBAT_NODENAME}').

main([]) ->
    help();
main([Arg | _]) ->
    Action = arg_to_action(Arg),
    case Action of
        preview ->
            preview();
        reload ->
            reload();
        commit ->
            commit();
        help ->
            help();
        unknown ->
            unknown(Arg)
    end.

arg_to_action("preview") ->
    preview;
arg_to_action("reload") ->
    reload;
arg_to_action("commit") ->
    commit;
arg_to_action("help") ->
    help;
arg_to_action(_) ->
    unknown.

help() ->
    Help = "Usage:~n"
        "        escript node_restructure.escript COMMAND~n"
        "~n"
        "  COMMANDS:~n"
        "    preview~n"
        "       Preview the structure based on current callback module.~n"
        "    reload~n"
        "       Reload the callback module in the Wombat VM and preview changes.~n"
        "    commit~n"
        "       Commit changes by restarting wombat.~n"
        "    help~n"
        "       Print this text.~n"
        "~n"
        "For further help, please consult the documentation!~n",
    io:format(Help),
    halt(1).

preview() ->
    connect(),
    PreviewData = call(wo_topo, get_preview, []),
    case PreviewData of
        [] ->
            io:format("No nodes are present in Wombat.~n");
        _ ->
            lists:foreach(
              fun({Family, Details}) ->
                      io:format("~p:~n", [Family]),
                      SortedDetails =
                          lists:sort(fun({N1, _, _}, {N2, _, _}) ->
                                             N1 =< N2
                                     end, Details),
                      [io:format("    ~p : ~p~n", [Name, DisplayName])
                       || {Name, _Id, DisplayName} <- SortedDetails]
              end,
              lists:sort(fun({F1, _}, {F2, _}) -> F1 =< F2 end, PreviewData))
    end.

reload() ->
    connect(),
    call(wo_core_master_app, load_alias_callback, []),
    io:format("Reloaded.~n").

commit() ->
    connect(),
    io:format("New nodes will be added according to new rules!~n"
              "Existing nodes would look like the following after committing:~n"),
    preview(),
    io:format("~n"
              "+-------------------------+~n"
              "|Wombat will be RESTARTED!|~n"
              "+-------------------------+~n"
             ),
    {ok, WombatCwd} = call(file, get_cwd, []),
    WombatConfig = filename:join([filename:dirname(WombatCwd),
                                  <<"files">>,
                                  <<"wombat.config">>]),

    IsSafe = proceed_on_clean_config(WombatConfig),
    case io:get_line("Are you sure to proceed? (yes/no) ") of
        "yes\n" ->
            modify_wombat_config(IsSafe, WombatConfig),
            io:format("Stopping Wombat... "),
            stop_wombat(),
            io:format("ok~n"),
            io:format("Starting Wombat... "),
            start_wombat(),
            io:format("ok~n"),
            io:format("Changes are committed.~n");
        _ ->
            io:format("Please type yes! Terminating!~n"),
            halt(1)
    end.

unknown(Command) ->
    io:format("Unknown command: ~p.~n", [Command]),
    help().

connect() ->
    case net_kernel:connect_node(?WOMBAT_NODE) of
        true ->
            ok;
        false ->
            io:format("Cannot connect to the Wombat node! Check Wombat logs for"
                      " errors!~n"),
            halt(1);
        ignored ->
            io:format("Cannot connect to the Wombat node! Make sure it is "
                      "running.~nThis script can only be used with a running "
                      "Wombat node.~n"),
            halt(1)
    end.

call(Module, Function, Arguments) ->
    case rpc:call(?WOMBAT_NODE, Module, Function, Arguments) of
        {badrpc, Reason} ->
            io:format("Error! Cannot run command (~p:~p(~p)) on remote node "
                      "with reason: ~p~n", [Module, Function, Arguments, Reason]),
            halt(1);
        Res ->
            Res
    end.

modify_wombat_config(true, _WombatConfig) ->
    %% Config option is already present.
    ok;
modify_wombat_config(missing, WombatConfig) ->
    %% We get the currently running wombat's CWD to compute the path of
    %% `wombat.config'. This is the safest way to get the correct file.
    case filelib:is_regular(WombatConfig) of
        true ->
            ok;
        false ->
            io:format("No wombat.config at ~s~n!", [WombatConfig])
    end,
    {ok, FD} = file:open(WombatConfig, [append]),
    io:format(FD, "~n~s~n",
              ["{set, wo_core, automatic_node_restructure_at_startup, true}."]),
    file:close(FD).

start_wombat() ->
    case filelib:is_regular("start.sh") of
        true ->
            %% This is a valid release, use that method
            os_cmd("./start.sh");
        false ->
            %% This is not a release, so maybe a development environment.
            os_cmd("rel/wombat/wombat/bin/wombat start")
    end.

stop_wombat() ->
    case filelib:is_regular("stop.sh") of
        true ->
            %% This is a valid release, use that method
            os_cmd("./stop.sh");
        false ->
            %% This is not a release, so maybe a development environment.
            os_cmd("rel/wombat/wombat/bin/wombat stop")
    end.

os_cmd(Cmd) ->
    Res = os:cmd(Cmd),
    io:format("~s~n", [Res]).

proceed_on_clean_config(WombatConfig) ->
    {ok, Opts} = file:consult(WombatConfig),
    OptionState = option_state(Opts),
    case OptionState of
        true ->
            %% Safe to proceed, because we will add nothing.
            true;
        false ->
            io:format("Automatic node restructure is set to false in "
                      "`wombat.config'.~n"
                      "Please set it to 'true' or delete the configuration.~n"
                      "Terminating without any modification.~n"),
            halt(1);
        missing ->
            %% Safe to proceed, because we will add the config option.
            io:format("The following config option will be added to the "
                      "`wombat.config':~n"
                      "   {set, wo_core, automatic_node_restructure_at_startup,"
                      " true}~n"),
            missing
    end.

option_state(Opts) ->
    case lists:keyfind(automatic_node_restructure_at_startup, 3, Opts) of
        %% Matching to explicit patterns can be used to check the validity of
        %% the setting.
        {set, wo_core, automatic_node_restructure_at_startup, true} ->
            true;
        {set, wo_core, automatic_node_restructure_at_startup, false} ->
            false;
        false ->
            missing
    end.
