# Wombat Plugin Architecture

## Motivations

### The original Wombat architecture

As Wombat is approaching its 1.0 release it became evident that the
initial architecture plans of the system need to be updated with the
lessons learned working on the beta versions. The architecture as of
0.7.0 looks like this:

<img alt="Wombat architecture" src="arch.png" width="600">

This approach decomposes the system to the following vertical layers
(in bottom-up order):

* __Agent code__ that executes on the managed node.
* __Middle manager__ the layer within Wombat that directly interfaces
  the managed node. For scalability reasons there can be multiple
  middle managers in a Wombat deployment.
* __Master__ the central processes of Wombat.
* __API__ the northbound interface of Wombat that allows access to the
  system for various user interfaces and higher level managers.
* __User interfaces__ the web dashboard and the CLI.

There is also a horizontal decomposition in the form of _plugins_. A
plugin adds certain functionality to the system (such as metrics
management or alarm handling). It typically spawns over all vertical
layers (such as metrics that contains user interface, REST API code,
master and middle manager processes and functions to execute on the
managed nodes as well). The plugin model is very flexible, the
interface towards Wombat's core is simply a pair of callbacks to start
the master and middle manager processes of the plugin. The plugin can
then subscribe to events that inform it about managed nodes being
added/removed/etc.

The original model is still very useful, however has a few problems:

* The _plugin_ concept of the architecture is different from the
  vision document's. The latter assumes a plugin manages a single OTP
  application on the node, but takes full responsibility for it by
  handling metrics, notifications and alarms at the same
  time. Decomposition of the code by OTP application is indeed
  desirable.
* There is no urgent need for further extending Wombat with orthogonal
  management areas (new plugins in the current architecture's
  sense). There is a common basic feature set offered by all Wombat
  installations:

  * Metrics
  * Notifications
  * Alarms
  * Orchestration (currently an experimental feature)
  * OTP application specific actions (which is only a planned feature
    at the time of writing)
* There are no frameworks, best practices or any other architectural
  guidelines for implementing the plugins (in neither sense of the
  word). This means a lot of duplicated code was written (e.g. for
  injecting modules or supervising processes in the managed nodes)
  which is hard to maintain.
* Depending on the managed node's OS type, ERTS and OTP application
  versions some features may not be available or need to be
  implemented in different ways. There are only ad-hoc solutions for
  this problem in various plugins.
* There are no common services offered by Wombat on the managed nodes
  for the plugins. Within the Wombat cluster at least some utilities
  can be found in the `wo_utils` application, but not on the managed
  nodes themselves. Although there's need for a number of such
  services, such as process supervision, code purging, addressing
  processes on Wombat etc.
* It turned out that there are some expensive queries that multiple
  agents would have to execute on the managed node, such as
  periodically retrieving some information from all running
  processes. Or the same piece of information (e.g. amount of memory
  allocated) may be the input of both a metric and an alarm
  rule. These kind of synergies cannot be used for optimizations when
  plugins are this independent.
* The functionality is difficult to extend in Wombat's core: plugins
  only have a public API for the end user (mostly the REST
  dashboard). It is not possible to add e.g. a process that receives
  new metric values and after some processing creates new
  notifications or metrics from them without being familiar with and
  using the internal APIs of the plugins involved.
* The original architecture doesn't plan with central data processing
  processes as bottlenecks. There are either master processes that are
  responsible for global functionality and middle manager processes
  that need to deal with a random subset of the nodes. However, there
  is need to perform expensive node independent tasks like verifying
  all nodes within a node family have the same module versions
  loaded. The architecture doesn't support scaling these kind of tasks
  by distributing them to worker nodes.
* It is now evident that Wombat shall monitor its agent processes for
  resource consumption (mostly CPU and memory, maybe I/O too). This
  should be part of the standard Wombat library/framework available
  for custom agents.
* The current plugins are not configurable: they always inspect
  everything on the managed node they are capable of and run
  measurements at a predefined rate. We need user control over what to
  do on the managed node (so for example some measurements that don't
  scale well may be disabled on production systems), and we need to
  improve the automatic configuration as well to avoid duplicated
  measurements (e.g. we don't need to report SASL logs when lager is
  running and configured to capture them anyway).
* The dashboard was not prepared for handling OTP applications.

  * It would make sense to model applications within the topology
    service of Wombat, as a new level beneath nodes.
  * Filtering notifications, metrics and alarms by application shall
    be implemented consistently.
  * The current plugins do not have features to control the nodes
    (save a couple of limited ad-hoc solutions), so there is no screen
    space reserved for such controls at all.
  * Wombat needs well defined data types for information retrieved
    from nodes. For example some alarm sources provide formatted text
    descriptions for alarms, while others fill in this field with
    generic terms. This should be solved by a consistent type system
    that allows the UI to display the information properly.

### Francesco's vision

The _Wombat Agent Modules_ document from Francesco proposes a plugin
architecture for Wombat. It describes the architecture through four
views:

* __Information model__ provides functions and data types for
  describing the managed resource (an OTP application). It needs to
  have two components:

  * Management data, that does not depend on or affect the logical
    state of the managed resource, such as the list of metrics
    available.
  * Logical data, which represents the state of the resource, such as
    the actual value of the metrics.
* __Functional model__ specifies the functionality to be generally
  supported by the plugins (an API). The document proposes a very
  simple API for the agent modules over which Wombat can query the
  list of metrics (management data), the current metric values
  (logical data) together with a callback interface for reporting
  alarms and notifications towards Wombat. The agents also have an
  init and terminate function call pair for life cycle
  management.
* __Organizational model__ is based on the software architecture of
  the managed node. It is defined in the Wombat topology subsystem,
  and the document doesn't detail it any further.
* __Communication model__ is based on the protocols to operate across
  the network. In case of Wombat it is the Erlang distribution
  protocol.

The document also brings up a list of additional requirements to
address in the plugin architecture:

* Choosing the injected code's version based on the OS, ERTS and app
  versions. The availability of dtrace or other optional VM features
  may also affect Wombat's behavior.
* Purging the old code of agent modules.
* The necessity of northbound interfaces (REST, SNMP etc.).

The _Wombat Agent Modules_ correctly identifies the need for a generic
information and functional model that can be applied to all
plugins. However it doesn't define enough details in its present form,
and the proposed API is almost as generic as the current one. In this
stage of development we need to implement more reusable frameworks and
leave less gaps for the plugin creators to fill in.

## Naming convention cleanup

To solve the problem of the overloaded plugin concept and to define
some currently nameless components from this point the following
naming convention will be used:

* A Wombat __subsystem__ will mean the same thing that used to be a
  plugin in the original architecture.
* A Wombat __plugin__ will refer to code managing a single
  application.
* A Wombat __report__ will refer to code that executes within Wombat
  but processes the data collected from the managed nodes in order to
  generate additional data of similar type. One such report may
  inspect notifications and raise alarms upon detecting certain
  patterns. Or it may convert a gauge metric into a histogram.

## The plugin architecture's vision

This chapter describes the long term goals of the plugin
architecture. The next chapter identifies the part of it that shall
fit into the 1.0 timeframe.

### Information model

The core of the new plugin architecture is a common information model
(IM) that will hold all the data Wombat operates on. The IM has a well
defined internal API over which agents and reports can interact with
it and can have generic mappings to REST and other NBIs.

#### The MO tree

Information shall be organized into a tree of Managed Objects
(MOs). An MO can represent for example a node family, a node, an OTP
application, a metric series, an alarm, a notification or just a
generic resource used for groupping together other MOs in a logical
hierarchy.

An MO has the following properties:

* __Name__ that uniquely identifies the MO _among its siblings_.
* __Class__ that defines the type of the MO (e.g. application or
  metric) and serves as a callback module too. The MO interface is
  designed to be generic, but the callback modules are responsible for
  the actual implementation, so they can choose appropriate DB backend
  or enforce some validation rules when the user tries to edit the MO.
* __Id__ is an internally assigned unique identifier of the MO _within
  its class_. So it is possible to directly look up an MO if you know
  its ID and class, regardless of its position in the MO tree. On the
  other hand it is not possible if you only know the name and
  class. For example probably every node MO will have an application
  MO named `kernel` under it. Only if you know the entire _path_ of
  the MO (a list of names) can you look it up, step by step starting
  from the root MO.
* __Attributes__ which are basically arbitrary key-value pairs
  describing the MO.

  __Operational status__ is an attribute that represents the actual
  (inspected) status of the resource the MO models. It is special
  because of a number of reasons:
  * Every MO must have an operational status attribute.
  * Although the actual values it may take depend on the class, but
    there's one common with very special semantics: `not_present`. If
    the MO is `not_present` it means the underlying resource doesn't
    exist (e.g. a managed node is disconnected, an application is
    stopped, an alarm is cleared etc.).
  * The `not_present` operational status is inherited by every child
    of an MO (and recursively by every descendant too). This is simply
    because a parent-child relationship between MOs model a
    container-contained relationship between the underlying resources:
    if the container is not there, the contained elements cannot be
    present either.
  * The MOs must be able to keep track of their original operational
    status and the inherited `not_present` status. When they no longer
    inherit `not_present` they need to return to using their own
    status. (This is similar to how visibility of GUI widgets work: if
    you hide the parent, its children get hidden too. When you show
    the parent the children will apply their own visibility setting
    again.)
* __Children__ are the MOs under this one.

This is an example MO tree described with an Erlangish syntax. Only
the MO names, classes and attributes are shown.

```
root :: root ->
  "wombat 0.8.0" :: node_family ->
    wombat@127.0.0.1 :: node ->
      kernel :: application ->
        version = [3,0]

        total_memory :: metric ->
          value = 1243012
          desc  = {'$text$', <<"The total amount of memory used by the VM">>}
          type  = gauge
          unit  = {'$text$', <<"Byte">>}
          min   = 0
          max   = unbounded
        end

        %% ...
      end
      os_mon :: application ->
        %% ...
      end
    end
  end
  "wo_test 0.8.0" :: node_family ->
    wo_test1@127.0.0.1 :: node ->
      %% ...
    end
    wo_test2@127.0.0.1 :: node ->
      %% ...
    end
  end
end
```

### Type information in the MO tree

If every important piece of information is represented in the MO tree
it will become possible to access all of it via the same generic API:
an Erlang module for internal Wombat processes and a REST interface
for the UI. An update of an attribute would look identical regardless
of editing a metric or an alarm MO for example.

However to present the data for a human user we need some extra type
information:

* The MO class tells the UI what kind of view and control to use for
  displaying the MO. For example a metric shall be represented as a
  chart, an alarm as a control that shows severity, status and
  description and can be extended to show comments as well.
* In case of attribute values plain text values shall be
  marked. Wrapping these values into a `{'$text$', _}` tuple is
  recommended for this purpose. (In order to distinguish them from a
  similar looking tuple coming from directly the managed node for some
  strange reason we could also use an optional `{'$term$', _}`
  wrapper.)

An example where Wombat can benefit from marking plain text attribute
values is an alarm's description: when Wombat raises an alarm on its
own this field will be filled in with formatted text for the human
user. However Wombat also subscribes to the SASL alarm handler. Alarms
received via that subscription will have an arbitrary term in their
description field. Adding this extra type info will allow the UI to
properly display both type of descriptions.

### MO versions

Wombat shall keep not just the current but historic versions of the
MOs too. This will server multiple purposes:

* In case of metrics it's very natural that we want to keep the
  historic values too.
* In case of alarms the historic values will serve as an alarm log.
* For other type of data (e.g. process info collected from the nodes)
  it will allow the detailed inspection of the managed node's state at
  any given time, which may be an invaluable help in post-mortem
  debugging. (Think about it like having a mini crashdump file from
  every minute.)

Adding the time dimension means it is not only possible to navigate up
and down in the MO tree to parent and child MOs but also to step
forward and backward in time.

Every MO version will cover not only a single point in time, but a
time range between two updates of the MO. For example if a metric is
sampled every 2 minutes each version of the metric MO will cover a 2
minute interval.

There are a few design constraints for version tracking of MOs:

* Versions shall be created with a 1 second resolution. That is, if an
  MO is updated two times within a single second only one new version
  of it will bre created.
* Normally MOs are not deleted permanently, only their operational
  status is set to `not_present` when needed. Very old MO versions
  will have to be cleaned up from time to time by separate jobs.
* The class and name of MOs shall not change during their entire life
  cycle. However their internally assigned ID may change. The reason
  is that the ID is intended to be a key in the underlying DB backend
  to allow quick access to the MO. If an MO changes very frequently it
  may make sense to use different IDs for different versions of it.

  However given a class, an ID and an arbitrary timestamp the callback
  module must be able to find the correct version of the MO, even if
  it has a different ID for the given timestamp. But this operation is
  expected to take significantly longer than looking up with the ID
  that belongs to the given timestamp.

It is worth pointing out that because MOs are not deleted, the _shape
of the MO tree_ (which MOs are there and what children they have) is
the same at _every_ time point. Only the attributes of the MOs
(including their operational status) changes from one time to another,
and sometimes their internally assigned ID too.

### Versatile DB backends

The actual storage mechanism will be hidden by the MO class. This is
necessary because the storage requirements vary greately:

* Metrics need frequent updates to small amounts of data.
* Applications, nodes and many other MOs change rather infrequently
  but have a complex internal structure.
* Some values, like crashdump files may take up several GBs and have
  no structure (they are typical BLOBs).

## Functional model

### REST mappings for the MO DB

A generic mapping of the MO DB to the REST API of Wombat is
possible. Trivially, every MO can be represented as a JSON object and
an URI path may map to a path in the MO tree. Version selection could
be possible via a parameter encoded in the URI for example. GET, PUT
and DELETE requests would allow manipulation of a single MO in a
straightforward way.

However to fully replace the currently available REST API some
advanced query capabilities are needed:

* Query a subtree of an MO with optional filtering on:

  * depth of the MO in the subtree
  * MO class
  * operational status (the default should be to return only MOs that
    are present)
  * value of specific attributes (including case insensitive and maybe
    even regexp matching)
* Query all versions of an MO within a time range.
* For efficiency reasons it may be useful to selectively query only
  certain properties of the MO, not all of them.
* Result pagination is required for every GET request (regardless of
  query type or filtering). This requires:

  * An optional `limit` parameter.
  * An optional `from` parameter that contains a reference (class, id
    and timestamp) to the last MO already retrieved by the client.
  * That the results are always sorted when a `limit` or `from`
    parameter is used.
* To compile the list of metrics available on a node family level the
  dashboard needs a query that would return those metric MOs that
  exist under the same relative path to every node MO in a node family
  MO. However, this would be very difficult to implement on the
  backend. It makes more sense to simply query _every_ metric MO under
  the node family and calculate the intersection with Javascript code
  on the client side.

The above requirements could be implemented via parameters encoded in
the URI. A more generic solution that may also be familiar to a lot of
programmers would be to adapt XPath for the task.

* Node types should be root, mo, attribute and class (instead of root,
  element, text, attribute, namespace, processing instruction and
  comment in the XML world).
* The list of axes would be different, partially to cover the
  difference in node types (e.g. no `namespace` axis is necessary, but
  we need `class`) but most importantly to allow navigation between
  versions (`older`, `newer`, `older-or-self` and `newer-or-self`
  axes).
* By default filter for MOs that are present, allow manual filtering
  for `not_present` MOs too.
* Some new functions would be useful:
  * `present()`
  * `at(node-set, timestamp)`
  * `not-before(node-set, timestamp)`
  * `not-after(node-set, time)`
* Pagination and sorting can be covered outside of the XPath query
  with additional parameters.

The XPath-based approach would be however more difficult to implement.

### Updating the MO tree via templates

When an agent module wants to write metric values, notifications or
other data to the MO tree it will not issue direct MO manipulation
calls. That would be similar to writing HTML markup from a HTTP
handler directly to a TCP socket. Just like modern web frameworks
separate the business logic and the HTML template the Wombat framework
will allow separation of the monitoring code and MO templates.

An MO template describes a portion of the MO tree, but may contain
variables (as attribute values or MO names), loops and
conditionals. When an MO template is used the caller will supply a
dictionary that maps variable names to actual values. Some default
variable bindings may also come implicitly from the environment: for
example the path from the root MO to the currently managed node's MO
is implicitly available for every agent process.

An important feature of a template is the ability to use a _truncate_
flag together with an MO. Truncate means that any other child MOs the
MO currently has, but not explicitly listed in the template (after
evaluating loops and conditionals) shall be removed (their operational
status set to `not_present`). A typical use case for this feature is
pulling alarms from the SASL alarm handler: at startup the agent does
only know what alarms are present on the node at the moment. So it
will create these alarms as MOs and remove any other siblings (old
alarms that Wombat still knows about, but are no longer present on the
node).

An example of how templates are planned to be used is available in
[this gist](https://gist.github.com/dszoboszlay/42990373fca361e4f9de).

Ultimately templates themselves shall be represented as MOs too. This
way it will become possible to configure things like the severity of
an alarm or change the description of a metric.

### MO update contexts

Every MO update shall happen within a context that the underlying
class module could use for validation. Example contexts:

* User update: Wombat may perform role based authorization of the
  request, or may simply deny updating certain values (e.g. measured
  metric values).
* Agent initialization: an agent initializes a branch of the MO tree
  when it starts to manage a node. However, if the MOs already exist
  (e.g. when reconnecting to an already managed node) the user may
  have already customized these MOs and the defaults proposed by the
  agent shall not overwrite the user customizations.
* Agent update: new metric values, notifications or alarms pushed from
  the agent to Wombat. Restrictions to user updates do not apply in
  this context (e.g. metric values can be written too).

### MO update events

MOs shall emit an update event when they change. Wombat processes may
subscribe to such events.

### Jobs

A job is a new concept of the plugin architecture. So far every agent
started to measure whatever it was capable of on every managed node
with hard coded parameters (e.g. polling time). This is not flexible
enough.

In the new model plugins will create _job MOs_ that represent the
measurements they can perform (e.g. subscribe to system monitor events
or poll `disksup` for statistics in `os_mon`). The job parameters
(like polling time or whether to execute the job at all) can be
manipulated by the user via the standard REST API calls.

Wombat will inform the agent to start or stop jobs as needed. This
could be in response to user updates in the configuration or to
changes on the managed node (e.g. if `disksup` is stopped, we need to
terminate the job that polls it as well).

On the agent side there will be behaviours for typical job types that
facilitate implementation.

* __Active job__ is the most flexible scenario: the job will run in
  its own process and can act on its own schedule. There will be a
  `gen_server` like interface: `init/1`, `terminate/1` and
  `handle_info/2` should be enough. This is most useful when the
  measurement means to subscribe to some kind of messages (e.g. system
  monitor or Mnesia events).
* __Polling job__ could be as simple as an MFA callback. Wombat will
  take care of calling this function periodically which shall return
  an MO template with parameters to be applied on the MO DB.
* __Sys job__ means to install a debug function (with `sys:install/3`)
  to an OTP process (e.g. a `gen_server`) for monitoring. The job
  implementation will receive the events, but won't have to take care
  of setup and cleanup for example.
* __Trace job__ is a job using the trace BIFs to monitor certain
  function calls in the system. Once again, the setup and cleanup is
  the responsibility of the behaviour, not the actual job
  implementation.

Wombat shall also provide some common services for jobs, like:

* Supervision: the job needs to be restarted when crashed.
* Performance metrics: measure the CPU and memory load of executing
  the job. This is especially important because jobs run on the
  managed node.

One more important thing is that job MOs aren't necessarily mapped
one-to-one to jobs running on the managed node. For example if there's
a job MO for iterating over all TCP sockets and collecting some
metrics on them and an other one for iterating over UDP sockets the
agent is free to choose to merge these into a single job (when both
started) because technically it will have to iterate over every _port_
in both cases, and it's much more efficient to do it only once. An
other example is when a clever agent offers two job MOs for the same
task, but with different tradeoffs: one may use polling (safe but
slow), the other tracing (more efficient, but may not work on every
managed node). The agent may only execute one job, even if the user
attempts to start both.

### Reports

A report is very similar to a job, however it runs on the Wombat node
and processes data in the MO DB. Reports may be responsible for:

* Raising alarms when metrics cross predefined thresholds.
* Calculating new metrics from existing ones (e.g. if you need the
  "memory usage without the code and atom memory" metric, or if you
  want to do statistical analysis on a gauge turning it into a
  histogram).
* Aggregating old metric values to form 5 min or 15 min averages.
* Generating multi-node (e.g. node family level) metrics and alarms.

The goal is to keep the agents simple and avoid expensive calculations
on the managed nodes. Therefore the agent jobs shall push raw data to
Wombat (e.g. various memory usage statistics) and let the reports
transform these to actual metrics and alarms (e.g. aggregate them,
check for crossing thresholds and so on).

### Managed node life cycle

Currently there are two states for nodes managed in Wombat: _up_ and
_down_ (orchestration adds a couple of additional states, but
basically it refines the down state only). This needs to be detailed a
bit further in the new architecture.

1. Wombat shall first inject the generic agent code and retrieve some
   information from the node:

   * OS type.
   * ERTS and running application versions.
   * The current incarnation of the node.
2. Wombat shall select the appropriate plugins for the node based on
   this information and inject their agent modules.
3. The set of jobs to execute on the node needs to be determined:

   1. Each plugin may inspect the node and create default job
      configurations.
   2. Plugins may override each other's configurations when necessary
      (e.g. the lager plugin may disable the SASL plugin's alarm
      reporting job if lager handles SASL logs on the managed node).
   3. Finally, the user's customizations are applied on the
      configuration taking precedence over the previous steps.
4. The node is now up and managed.

   * Jobs that were enabled in the previous step and are always on are
     started.
   * Jobs can be started and stopped any time from now on.
   * Wombat shall monitor the node for new applications being started
     and stopped. Each new application can be initialized according to
     the above strategy.
5. If the node is disconnected from Wombat, the generic agent services
   shall take care of:

   * Deciding how long to wait for the connection to be restored
     before unmanaging the node.
   * Buffering the messages from the node to Wombat, if the agent is
     configured so.
   * Resending these messages once the connection is restored.

6. When the node is unmanaged the agent needs to clean up after
   itself.

The Wombat agent may become available as a stand-alone application
that can be integrated into Erlang releases. For such scenarios the
life cycle model needs to deal with:

* Taking over control of a running agent in the node.
* Allowing init to take control of an injected agent (in case Wombat
  injects and starts it before the bootscript would start the embedded
  one).
* Handling custom agents present on the node: they shall be included
  in the job configuration process, including their custom overrides.
* Distinguishing between injected and embedded agents: the latter ones
  shall be left there after unmanaging the node.

## Organizational model

The organizational model is encoded in the MO tree in the new plugin
architecture. We will have node families, nodes and applications on
the top three levels. Probably a separate branch for internal Wombat
configuration would be useful too (e.g. the MO templates).

## Communication model

The main channel remains the Erlang distribution protocol. We need
this in order to inject the agent into the managed node. However on
the long term we will need some other side channel too for transfering
big bulks of data (like crashdumps).

The middle manager nodes shall be also extended sooner or later to use
a customised distribution protocol for bridging differences in:

* short and long name usage
* tick time settings
* use of SSL
* addressing issues (consider a middle manager running on a gateway
  host to an internal network: the managed nodes may have IPs that
  cannot be accessed from anywhere else; in worst case there may even
  be two separate `node@10.0.0.3` nodes managed by Wombat via
  different middle managers)

Ultimately we shall make Wombat partitionable. Currently as Wombat
connects to all nodes with distribution protocol it is possible to
have full access to any managed node from any other managed node. This
is a security risk, and some organisations may not allow to connect
their production nodes to staging or development nodes this way.

In order to overcome this problem we will have to allow connecting a
middle manager to Wombat via a non-distribution link. It could be a
TCP or REST channel, a RabbitMQ link or even the use of a shared,
external DB.

## Agent services

The following services are needed by most agent processes. They shall
be supported with libraries injected into the managed node.

* Supervision of agent processes.
* Life cycle management that covers:
  * Bootstrapping the node (retrieving OS and running application
    versions, node incarnation).
  * Tracking basic changes (e.g. applications being started/stopped,
    init phases).
  * Forwarding job control requests to the agents (e.g. live metrics
    require more frequent updates on statistics).
  * Handling connectivity issues (local caching of data, reconnect
    attempts and timeout etc.).
  * Cleaning up the node (stopping jobs, processes and purging
    dynamically loaded modules).
* Code loading and purging.
* Metrics for the load generated by Wombat agents.
* Communicating with Wombat. This involves:
  * Addressing processes with symbolic names (not PIDs that may change
    over time).
  * Local caching of the data when Wombat is disconnected.

## Customizable UI

The UI shall be more customizable in the future. By representing
dashboards, control forms and similar UI elements as MOs it will
become possible to achieve this goal too.

The plugins shall also allow the users to perform management
operations on the nodes (e.g. after a Mnesia netsplit select a master
node and restart Mnesia on disconnected nodes to resolve the
problem). These actions will need html forms or even more
sophisticated controls (e.g. a netsplit visualiser implemented with
Javascript code) as a user interface. The UI should be able to render
these controls dynamically, based on their respective MOs.

## Wombat clustering

As the plugin architecture introduces the concept of reports and tries
to push the compute intensive tasks to the Wombat node itself, it is
evident that we will need to move away from the 1 master node, many
middle managers concpet.

Middle managers shall be responsible solely for connectivity towards
the managed nodes. The master node shall be replaced with a cluster of
Wombat core nodes that execute the reporting tasks. It may be easyer
to keep a dedicated master node nevertheless that cordinates the rest
of the core nodes of course, but distribution is a must have on the
long term.

# The 1.0 scope

Wombat 1.0 is planned to include the following subset of the plugin
architecture:

* The MO DB.
* Wrappers for the existing notification, alarm and metric DBs that
  present the data in the new MO model too.
* REST API for accessing the MOs.
* The UI using the new REST API.
* Agent services without some advanced features like:
  * Buffering updates while Wombat is disconnected.
  * Ability to embed the agent into a release (instead of injecting it
    on the fly).
* At least the basic active and polling job behaviours.
* Node lifecycle management in Wombat, without the user customisations.
* At least one plugin implemented in the new framework (POC).

Basic things that will be most likely left out:

* Reports, if not necessary for the POC plugin.
* Migrating the existing metrics, notifications and alarms agents to
  the new architecture.
