# Contract between Wombat Agent and Plugin (for WombatOAM 1.0) #

This document is the result of discussions in the Erlang Solutions
Budapest office between Torben Hoffman, Daniel Szoboszlay and Robby
Raschke on Tue, 12. August 2014.

The original discussion document can be found in
https://gist.github.com/rtraschke/f1a97fca30b2c477a304 .


## Introduction ##

A Wombat *plugin agent* is the code injected by Wombat into a *managed
node* that is responsible for generating streams of metrics,
notifications, and alarms arising from a particular *functional
part* of the managed node. The immediate consumer of these streams
is the *wombat agent*, which is also injected by Wombat into the
managed node and is in charge of communicating interesting information
about the managed node to Wombat. The initial assumption is that
the functional part of the managed node triggering metrics,
notifications and alarms maps one-to-one onto Erlang *applications*.


## Plugin Capabilities ##

Collectively, the metrics, notifications and alarms that a plugin
may generate are called its *capabilities*. The wombat agent needs
to be able to query a plugin's capabilities in order to tell wombat
about the possible data it may receive when this plugin is activated
in the managed node. These capabilities are identified via a
hierarchical naming scheme and they state what type of capability
they provide (metric, notification or alarm) and further metadata
that allows Wombat and its user interface to manage these capabilities.


### Capability Metadata ###

Using Erlang type language (apologies in advance for any syntactical
errors), the overall capabilities of a plugin are defined as a list
of pairs containing the id and the definition (or info) of the
capability (as a proplist):

```erlang
-type capabilites() :: list( capability() ).
-type capability() :: { capability_id(), list( capability_info() ) }.
```

Since the naming scheme for capabilities is required to be hierarchical,
a capability id is a list of UTF-8 encoded binary strings:

```erlang
-type capability_id() :: list( binary() ).
```

The intention here is that the capabilities of all the plugins
together provide a tree of *managed objects*. It is most likely
that the main initial branches are all named after the Erlang
applications from which they are gathering information. For example,
we may have a plugin for the kernel application would gather metrics
with ids like `[ "kernel", "total_memory" ]` and `[ "kernel",
"process_count" ]`, as well as alerts with ids like `[ "kernel",
"process_limit" ]` and `[ "kernel", "atom_limit" ]`. Whereas a
plugin that gathers information about the OS of the system where
the managed node is running, might have metrics like `[ "os_mon",
"cpu_sup", "cpu_avg1" ]` and `[ "os_mon", "memsup",
"total_memory" ]`. This would lead to a tree structure like this:

* kernel
    * total_memory
    * process_count
    * process_limit
    * atom_limit
* os_mon
    * cpu_sup
        * cpu_avg1
    * memsup
        * total_memory

The information that describes the capability, i.e., its metadata,
is a proplist of atom keys and associated values (hopefully, the
key names used below are self-explanatory):

```erlang
-type capability_info() :: metric_info() | notification_info() | alarm_info().

-type metric_info() ::
      { 'type', 'metric' }
    | { 'description', string() }.
    | { 'metric_type', 'gauge' | 'counter' | 'histogram' | 'spiral' | ... }
    | { 'metric_default_freq', millisecs() }
    | ...
    .

-type notification_info() ::
      { 'type', 'notification' }
    | { 'description', string() }
    | ...
    .

-type alarm_info() ::
      { 'type', 'alarm' }
    | { 'description', string() }.
    | { 'alarm_severity', 'minor' | 'major' | 'urgent' }
    | ...
    .

-type millisecs() :: non_neg_integer().
```

*The list of all required capability metadata values needs to be
completed.* The information carried in this metadata must allow the
agent and Wombat to handle the incoming data from the plugin.


### Capability Data ###

To be completed.

```erlang
-type capability_data() :: term().
```

> So far, this defines the metadata of metrics, notifications and
> alarms. In terms of standard management nomenclature, this covers
> part of the *management data* of the *information model*. We also
> need to define the *logical data*, that is, define the values that
> a plugin will send to the agent for particular types of capabilities.


## Interface ##

A plugin must adhere to the following interface in order to interact
with the wombat agent. This interface is divided by the simple
lifecycle of a plugin: start and initialisation, operation, and
termination. The operation phase of the lifecycle must allow the
querying of supported capabilities as well as activation and
deactivation of capabilities.

> Do we want to "enforce" this interface via a behaviour?

### Initialisation ###

The plugin initialisation call is expected to start the plugin process.

```erlang
-spec start_link(Wombat_Agent_Pid :: pid(), ...) -> {'ok', pid()} | {'error', Reason :: term()}.
```

> Do we want to mandate something similarly well defined as a
> gen_server here or keep things light?

> We need to define certain error reasons that signal particular
> expected types of initialisation failure to the agent.


### Termination ###

The plugin termination call is expected to clean up any "stuff"
(internal state, child processes, event handlers, etc.) and stop
the plugin process. After termination, the plugin module (and
registered dependency modules) must be able to be completely removed
from the managed node using `code:delete` and `code:soft_purge`.

```erlang
-spec stop(Plugin_Pid :: pid(), ...) -> ok.
```


### Operation ###

During the lifetime of the running plugin, its capabilities may be
queried and capabilities may be turned on or off. When turning on
metrics, it must be possible to override the default metric gathering
frequency.

```erlang
-spec capabilities(pid()) -> capabilites().

-spec activate(pid(), list( Activation )) -> 'ok' | { 'error', Reason :: term() }
    when Activation :: capability_id() | { capability_id(), millisecs() }.

-spec deactivate(pid(), list( capability_id() )) -> 'ok' | { 'error', Reason :: term() }.
```

> If more complicated configuration of a capability at activation
> time is required, then a dedicated type must be created.

A plugin must only collect data for metrics, notifications or alarms
that have been activated. The wombat agent provides a function to
the plugins to report values.

```erlang
-spec wombat_agent:report(Wombat_Agent_Pid :: pid(), capability_id(), term()) -> ok.
```


## Example ##

An example for a metric plugin.

*Note:* This is uncompiled and therefore full of syntax errors and
bugs. It is also not really a stunning example in terms of the
processes and triggering mechanisms. There is also no checking for
valid IDs getting passed in. Apologies, it was created under time
pressure.

It is very likely that some of the code will become available as a
small library for assisting plugin writers. For example, the code
to manage the activation/deactivation is going to be pretty common
across plugins.

```erlang
-module(port_info).

-behaviour(gen_server).
-behaviour(wombat_plugin).

-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
    terminate/2, code_change/3]).
-export([start_link/1, capabilities/1, activate/2, deactivate/2,
    stop/1]).

%%%=============================================================================
%%% Wombat Plugin API
%%%=============================================================================

start_link(Wombat_Agent_Pid) ->
    gen_server:start_link(?MODULE, Wombat_Agent_Pid, []).

% Completely static capabilities, no call into server required.
-define(DEFAULT_FREQUENCY, 5 * 1000) % every five seconds
capabilites(_Pid) ->
    [
        { [<<"port_info">>, <<"open_ports">>], [
            {type, metric},
            {description, <<"Number of open ports">>},
            {metric_type, gauge},
            {metric_default_freq, ?DEFAULT_FREQUENCY}
        ]},
        { [<<"port_info">>, <<"locking_ports">>], [
            {type, metric},
            {description, <<"Number of locking ports">>},
            {metric_type, gauge},
            {metric_default_freq, ?DEFAULT_FREQUENCY}
        ]},
        { [<<"port_info">>, <<"port_output">>], [
            {type, metric},
            {description, <<"Alive ports total output in bytes">>},
            {metric_type, counter},
            {metric_default_freq, ?DEFAULT_FREQUENCY}
        ]},
        { [<<"port_info">>, <<"port_input">>], [
            {type, metric},
            {description, <<"Alive ports total input in bytes">>},
            {metric_type, counter},
            {metric_default_freq, ?DEFAULT_FREQUENCY}
        ]}
    ].

activate(Pid, Capabilites) when is_list(Capabilites) ->
    gen_server:call(Pid, {activate, Capabilites}).

deactivate(Pid, Capability_Ids) when is_list(Capability_Ids) ->
    gen_server:call(Pid, {deactivate, Capability_Ids}).

stop(Pid) ->
    gen_server:cast(Pid, stop).

%%%=============================================================================
%%% gen_server callbacks
%%%=============================================================================

-record(state, {
    wa_pid,
    sub_proc_ids = [] :: {capability_id(), pid()}
}).

init(WA_Pid) ->
    {ok, #state{}}.

handle_call({activate, Capabilites}, _From, #state{wa_pid = WA_Pid, sub_proc_ids = Pids}) ->
    Updated_Pids = activate_capabilities(Capabilites, WA_Pid, Pids),
    {reply, ok, #state{sub_proc_ids = Updated_Pids};
handle_call({deactivate, Capability_Ids}, _From, #state{sub_proc_ids = Pids}) ->
    Remaining_Pids = deactivate_capabilities(Capability_Ids, Pids),
    {reply, ok, #state{sub_proc_ids = Remaining_Pids}.

handle_cast(stop, State) ->
    {stop, normal, State}.

handle_info(_Info, State) ->
    {noreply, State}.

terminate(_Reason, #state{sub_proc_ids = Pids}) ->
    [exit(Pid, kill) || {_Id, Pid} <- Pids], % no need to be polite, just terminate them
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%=============================================================================
%%% Functionality Implementation
%%%=============================================================================

activate_capabilities([], _WA_Pid, Pids) ->
    Pids;
activate_capabilities([Capability | Rest], WA_Pid, Pids) ->
    {Id, Pid} = activate_capability(WA_Pid, Capability),
    activate_capabilities(Rest, WA_Pid, [{Id, Pid} | Pids]).

activate_capability(WA_Pid, {Id, Frequency}) ->
    Pid = spawn_link(?MODULE, metric_loop, [WA_Pid, Id, Frequency]),
    {Id, Pid};
activate_capability(WA_Pid, Id) ->
    activate_capability(WA_Pid, {Id, ?DEFAULT_FREQUENCY}).

deactivate_capabilities([], Pids) ->
    Pids;
deactivate_capabilities([Id | Rest], Pids) ->
    case lists:keytake(Id, 1, Pids) of
        false ->
            deactivate_capabilities(Rest, Pids);
        {value, {Id, Pid}, Remaining_Pids} ->
            exit(Pid, kill),
            deactivate_capabilities(Rest, Remaining_Pids)
    end.

metric_loop(WA_Pid, Id, Frequency) ->
    timer:sleep(Frequency),
    report_metric(WA_Pid, Id),
    metric_loop(Id, Frequency).

report_metric(WA_Pid, Id = [<<"port_info">>, <<"open_ports">>]) ->
    Metric_Value = length(erlang:ports()),
    wombat_agent:report(WA_Pid, Id, Metric_Value);
report_metric(WA_Pid, Id = [<<"port_info">>, <<"locking_ports">>]) ->
    Metric_Value = lists:foldl(
            fun(P, A) ->
                case erlang:port_info(P, locking) of
                    {locking, false} ->
                        A;
                    {locking, _} ->
                        A + 1;
                    undefined ->
                        A
                end
            end, 0, erlang:ports()),
    wombat_agent:report(WA_Pid, Id, Metric_Value);
report_metric(WA_Pid, Id = [<<"port_info">>, <<"port_output">>]) ->
    Metric_Value = ists:foldl(
            fun(P, A) ->
                case erlang:port_info(P, input) of
                    {input, Input} ->
                        A + Input;
                    undefined ->
                        A
                end
            end, 0, erlang:ports()),
    wombat_agent:report(WA_Pid, Id, Metric_Value);
report_metric(WA_Pid, Id = [<<"port_info">>, <<"port_input">>]) ->
    Metric_Value = lists:foldl(
            fun(P, A) ->
                case erlang:port_info(P, output) of
                    {output, Output} ->
                        A + Output;
                    undefined ->
                        A
                end
            end, 0, erlang:ports()),
    wombat_agent:report(WA_Pid, Id, Metric_Value).
```
