Plugins Design
==============

Goals
-----

The plugin architecture intends to provide a common structure for creating agent code/plugins to run on the managed nodes in order to provide monitoring information to Wombat, not already collected by Wombat itself.

It is important to specify what we want to achieve by creating this plugin infrastructure. The requirements and business needs are not completely clear or detailed enough.

What other kinds of plugins/functionalities we want to provide in Wombat in the future besides the alarms, notifications, logs and metrics? This is important and might influence the design of the plugins.

We should be able to eventually rewrite the current agents as plugins following the new format in order to replace them.

Design decisions taken
----------------------

We want to use a PUSH communication model between the plugins and Wombat. Wombat plugins push information to a process which relays it to Wombat (“Plugin Relay” on the node, towards `wo_plugins_node_dispatcher` in Wombat). This model fits the current implementation of the logs (`logs_mm`) and the alarms (`alarms_mm`), but not that of the metrics (`metrics_mgr and metrics_live`), since those are implemented following a pull model.

The message format will be tagged tuples, and it should be possible to add new messages in the future beyond those created to report metrics, alarms and logs.

The plugin support shall not require extensive changes to the current architecture or functioning of Wombat. The plugins shall use the existing data paths in the system (the middle manager processes).
Open questions

We will need a way of activating/deactivating certain information generated at the plugins (capabilities activation/deactivation). 

We will need to provide a way for Wombat to make a call to the plugins (like for example we might need to allow Wombat to acknowledge an alarm on the Elarm server running on the managed node).


Architecture
------------

### Current functionality/architecture

At the moment we have four agents implemented on Wombat: `wo_alarms_agent`, `wo_logs_lager_agent`, `wo_logs_elogger_agent` and `wo_metrics_agent`. From this, alarms and logs spawn processes on the managed node that then push through a `gen_server:cast()` information into the Wombat Middle Manager processes (`wo_alarms_mm.erl` and `wo_logs_mm.erl`). For the metrics however, there are two middle managers (`wo_metrics_node_live` and `wo_metrics_node_mgr`) which pull informations through RPC calls from the managed node, which has the `wo_metrics_agent` code injected but does not have any process running. This difference makes it more complicated to integrate the metrics functionality with our plugins since we want to use a uniform push model.

### Proposed Architecture

The plugins support is added to Wombat in its own application (`wo_plugins`). Its supervisor tree is started under the middle manager tree similarly to other existing Wombat subsystems.

The plugin subsystem will be delivering data to the other ones, by sending messages to the alarm and logging middle managers. This ensures that data is collected and presented in a similar fashion for built-in agents and customer-written plugins alike.

### Wombat Middle-manager processes

The `wo_plugins_controller` process (a `gen_server`) is subscribed to node events (starts, stops) and manages the processes that will be responsible for managing plugins for the newly added managed nodes.

For each managed node two processes are started:

- `wo_plugins_node_mgr` which is responsible for injecting the supervisor hierarchy and start the required plugins on the node,
- `wo_plugins_node_dispatcher` which is the receiver of data pushed to Wombat by the plugins, it is responsible for delivering those messages to the appropriate Wombat subsystems.

### Supervision tree on Wombat side

![Supervision tree on Wombat side](./wo_pl_sup-tree-w.png)

### Processes on the managed node

The plugins injected and started on the managed nodes need supervision and support processes providing the communication channel towards Wombat.

The “Agent Controller” will be the peer of the wo_plugins_node_mgr process on the Wombat side. It is responsible for monitoring the connection towards Wombat and starting the plugins.

The process labelled “Plugin Relay” is the one responsible for providing the capability to push data to Wombat from the plugins — the opaque `WRef` passed to plugins at startup will contain its pid.

The injected hierarchy needs to be stopped and the code purged when connection to Wombat is lost or the node is removed from Wombat. Special care should be taken not to leave processes running on unmanaged nodes. This also assumes that plugins are well-behaved (e.g. are proper gen_servers, not starting other processes, etc), but we have no way to enforce this.

### Supervision tree on the managed node

![Supervision tree on the managed node](./wo_pl_sup-tree-mn.png)

### Startup Sequence on the Managed Node

On start, the supervisor tree will start the Agent Controller (`wombat_plugin_controller`) and the supervisor for plugins. When calling `connect/2` it starts the relay process.

### API on the managed node

`wombat_plugin_controller`: Process that handles the bidirectional communication towards Wombat. Wombat will only make calls to this API on the managed node. This process is registered by name on the managed node.

- `connect(NodeManager, NodeDispatcher)`  Establishes connection towards Wombat and provides the controller with the necessary pids on the Wombat side. It then starts the plugin relay process.
- `disconnect()` Stop all processes and purge all code.
- `which_applications()` This can be useful information for the Wombat side regarding which plugins to start on the managed node.
- `start_plugin(Name)` It is assumed that the plugin’s code was already injected by Wombat when this function is called.
- `stop_plugin(Name)`
- `started_plugins()`

`wombat_plugin` (the communications relay and main user API): `gen_event` or `gen_server`, handles data push to Wombat:
- `start_link(DispatcherPid)`
- `report(Ref, Data)`

### Starting the supervisor tree on the managed node and code cleanup

When starting the `wombat_main_sup` on the managed node we have
thought about two possible approaches. The first one (the one used in
the initial implementation), starts a process that doesn’t do anything
but to start the main supervisor under it. The other possible approach
(that we switched over later) it is to write an application behaviour
so that the main supervisor can be started as an application. It is
possible to first inject all the code to the managed node and then
load and start the application (there are functions for that on the
application module). This has the advantage that we are following OTP
(starting the main supervisor correctly), we can use restart
strategies for the application (i.e. if the whole supervisor tree goes
down, if the app gets restarted it could potentially then do a code
cleanup) and the plugins are also terminated correctly when the
managed node is stopped. This also plays well with the possibility of
having the whole code running on the managed node as an application
that can be added to a release.

When Wombat connects to a manage node it assumes to find the
`wombat_plugin` application in one of the following states:

- Injected scenario: the application is not present on the node. All
  the modules and the application definition are loaded by Wombat, and
  the application will destroy them all upon termination.

- Embedded scenario: the correct version of the application is loaded
  and started on the node. Wombat will simply try to connect to
  it. When the connection to Wombat is terminated only the plugins
  (processes and injected code) are removed, the application is left
  running.

- On demand scenario: the correct version of the application is loaded
  but not started on the managed node. Wombat will start it, and when
  the connection is terminated the application will stop itself (after
  removing the plugins).

- An incorrect version of the application (different from what Wombat
  has) is loaded and/or started. In this case we won't manage the
  node.

Other edge case scenarios (e.g. some modules are loaded on the managed
node, but the application itself is not) are ignored.

The `wombat_plugin_controller` is responsible for detecting the
termination of the connection towards Wombat and cleaning up the
plugins at this point. Unless the application was started in embedded
mode it also calls `application:stop/1`.

The clean up of the injected application and the base application
modules in injected mode is the responsibility of the application stop
hook.

#### Supervision tree on the managed node — Module names
 
![Supervision tree on the managed node](./wo_pl_sup-tree-mn-named.png)
 
### Plugins interface

A user-created plugin for WombatOAM shall be contained in a single module implementing the gen_server behaviour. The BEAM code of the plugin module is stored in a file accessible by Wombat and will be injected into the managed node when it's added to Wombat.

It's important to realise that the plugin code will run on the managed node under the supervisor hierarchy started by the Wombat agent. As a consequence, only the presence of the basic Erlang system can be assumed, and the functions provided by the wombat_agent API module can be used to communicate with Wombat:

See also https://github.com/esl/wombat-docs/blob/master/plugins.md

#### Further constraints and notes on the plugin module implementation

- The `init/1` function will be called as `init([WRef])`
- `terminate/2` can be used cleanup if necessary (the `gen_server` process will need to trap exits then)
- Wombat won't call `handle_call/2`, `handle_cast/2` or `handle_info/2`, it's solely the plugin's responsibility to schedule its work and push data to the Wombat agent. – _This may change in the future versions, when plugins might require bidirectional communication with Wombat_
- Plugin modules shall be named `wombat_plugin_<APPNAME>.beam` and be placed in a specific directory


## Implementation steps

### Steps to deliver the plugin infrastructure for the 1.0 release

Design the plugin interface correctly and ensuring that it can be extended to provide all the functionalities that we need in order to reimplement the current agents (metrics, alarms, logs) and that we might need for future agents/plugins or functionalities.

Design and implement the new modules/structure in the current Wombat code. Make sure it is possible to write plugins, run them and that the pushed information gets into Wombat.

Integrate this new infrastructure with the current alarms, notifications and logs MMs in order to be able to push logs and alarms from the new plugins infrastructure. This is a not difficult change.

Further work and architectural consolidation

Rewrite the current logs and alarms agents into the new plugin format, in order to replace the two current agents with a set of dedicated plugins.

Following tasks might be reimplementing the alarms agents and MMs to the new plugin format, to modify the UI and Wombat itself in order to take advantage of plugin functionality such as activating/deactivating reports (capabilities) and implementation of new plugin functionalities beyond alarms, notifications and logs.

# Risks

At least basic UI support needs to be provided, e.g filtering and displaying alarms and notifications by originating plugin

Adapting the current metrics functionalities to the new plugin format is a big and complicated change. To adapt the current metrics “middle manager” processes to accept metrics notifications from the plugins seems to us to be complicated, will be a workaround/hack anyway and will not add any value. The way of doing it correctly would be to write a metrics plugin using the new format, then write a new MM from scratch following the push model and once this is done and tested, replace the current MMs and agent with the new ones.

Injected code: At the moment Wombat injects the three agents code into the manage node. It also runs two processes in it, one for the alarms agent, one for the logs. The metrics code gets called through rpc by the Wombat side and therefore there is not running process on the managed node. Having the new plugin infrastructure we have a whole supervision tree running on the managed node, which means many processes running there, and more code to inject. Also, it might happen that we use libraries on the plugins code that are not injected/available on the managed node, causing runtime errors. A possible better approach might be to pack as an application the supervision tree of plugins running on the managed node, give that to customers so that they can add to the release. Then it will get compiled together with the client code, the client will be able to run their tests with a release that includes the Wombat code. My question is, if you were a client, would you allow an external system, which is not open source and therefore you cannot see what it does, to inject code into your nodes running in production?


# Delivery

## Goal for the 5th of September — Wombat 1.0

To have a functioning process structure (diagrams above) and plugins skeletons/callbacks defined.

Be able to send alarms, notifications and log messages which will be able to get displayed into Wombat UI as is is right now without complex changes to it.

UI work might be needed in order to display notifications/alarms on a per-plugin basis.

## Further goals

Add metrics pushing functionality to plugins.

Allow selectively switch on or off particular functions of a plugin (“capabilities”).


