# Infrastructure of tooling

Recently, a bunch of new feature requests about tooling has been received.
Wombat is trending to become an active agent. That means users can command
Wombat to perform such actions on the managed nodes that is likely to change the
behaviour of the system and cause side-effects. Introducing these features with
the current architecture of Wombat wouldn't be a perfect choice as Wombat was
prepared for different tasks. Also, introducing each feature with an ad-hoc
approach shall be avoided as to keep the relatively good code quality, the
easily maintainability and extensibility. So, it is good time to design the core
architecture of these features.

My methodology is to analyse as many to be added features as possible to ensure
that the infrastructure will be able to work with all. I will extract the common
properties and also form some categories. Further investigation then will reveal
the pre-requisites, the risks and the guarantees. That will allow me to design
the architecture and test it using dynamic models.

## Requirements

The architecture shall meet the followings.

- Easily extendible.
- Minimise GUI tasks.
- Tailor the set of features to any environments (managed nodes).
- Allow users to add their simple plugins.
- Plugins and Wombat shall remain loosely coupled.
- Wombat shall not know which plugin will do the task.

## Main features to be added

I analysed the following features to get the broad picture:

- Change configurations.
- Soft purge modules.
- Kill the shell.
- Force GC.
- Etop.
- Observer.
- User commands to be executed on a managed node / node family.
- (tooling)

## Common things

I have found the following statements hold for all the features.

- Started only on demand.
- Running only for a short period of time.
- The result is not persisted.
- the power of each feature varies among the managed nodes, can only be
determined runtime.
- The full set of a feature may be organised into multiple plugins.
- The possible arguments can be passed to each feature vary among the managed
nodes.
- Two or more features can cover the same functionality.


## Categories

### Command vs stream of temporal data Commands are executed on the managed node
as atomic non-interruptible operations. Their essence is the side effect caused
on the managed node. Certain commands may require exclusive execution. That
means while the command is being executed no other command can be started. They
can fail or succeed. The result is always returned.

A stream of temporal data means a series of homogeneous data that is
periodically pushed to Wombat. All the elements share the same data structure
but their values can differ. Every value is the result of calling the same
function that somehow measures a subject. Concurrent streams are allowed,
however, the number of the exact duplicates should be minimised to avoid
unnecessary overhead on the managed node. A stream is initiated by a special
command that trigger the corresponding plugin to start periodically pushing data
in question towards Wombat. The command must be well-formed and complete that
exactly defines how the subject shall be measured. The stream can be terminated
by another command that exactly identifies the stream to be closed. Of course,
any unexpected failure also terminates the stream. Closing streams shall always
succeed whilst initiations can fail. Failures shall be reported to Wombat.

### Scratching commands
Write some examples to roughly model some random, but concrete commands.

- Change configuration. Set mnesia's dump_log_time_threshold to 5000 can be written as
  ```
  {command,
   set_config,
   [{application, mnesia},
    {key, dump_log_time_threshold},
    {value, 5000}]}.
  ```
  We expect two results: `ok | {error, Reason}`.
- Allow users to soft purge modules to get rid of old code. Calling soft
 purge for modules having two versions loaded in the VM is not a risky
 operation. We need no parameter. It can be executed by separate
 processes at the same time.
  ```
  {command, soft_purge_old_code, []}.
  ```
  We expect that the result will be ok | {error, Reason}.
  It will be ok most of the time.
- Kill the shell.
  ```
  {command, kill_the_shell, []}
  ```
  We expect ok | {error, Reason} to be returned.
- Force GC.
Note that forcing GC on all the processes can be a long running task!
  ```
  {command, force_gc, []}.
  ```
  We expect ok | {error, Reason}.
- Execute user commands.
Users'd like to remotely execute their Erlang scripts even on every node of
a node family. That means arbitrary expressions shall be evaluated on the
managed node(s).
Observe three things. First the user's query may fail, second the user's
query can last for hours, third its value may be important for the user.
  ```
  {command, eval, [{expr, "lists:sort([1,2,3]) + 42"}]}.
  ```
  The result can be {error, Reason} or
  ```
    "{'EXIT',{badarith,[{erlang,'+',[[1,2,3],42],[]},
                   {erl_eval,do_apply,6,[{file,"erl_eval.erl"},{line,661}]},
                   {erl_eval,expr,5,[{file,"erl_eval.erl"},{line,434}]},
                   {shell,exprs,7,[{file,"shell.erl"},{line,684}]},
                   {shell,eval_exprs,7,[{file,"shell.erl"},{line,639}]},
                   {shell,eval_loop,3,[{file,"shell.erl"},{line,624}]}]}}".

    {ok, [{data_structure, value},
          {label, "Result"},
          {data, "{'EXIT',{badarith,[{erlang,'+',[[1,2,3],42],[]},
                   {erl_eval,do_apply,6,[{file,"erl_eval.erl"},{line,661}]},
                   {erl_eval,expr,5,[{file,"erl_eval.erl"},{line,434}]},
                   {shell,exprs,7,[{file,"shell.erl"},{line,684}]},
                   {shell,eval_exprs,7,[{file,"shell.erl"},{line,639}]},
                   {shell,eval_loop,3,[{file,"shell.erl"},{line,624}]}]}}"}]}.
  ```

### Example streams of temporal data

  - Etop. We want to get the table of those 10 processes consuming the most
  memory in each 5 seconds. We just want to list the processes.
  So, we open a stream.

        {stream, process_list, [{lines, 10}, {interval, 5}, {sort, memory}]}.

    The result tells us that the stream got initiated and the relevant data will
    be tagged by {etop, _StreamID = 1}. Data should be organised into a table
    that has the defined headers. No further interaction with the data can be
    expected.

        {ok,
        {etop, _StreamID = 1},
        [{data_structure, table},
        {type, static},
        {headers, ["Pid", "Name or Initial Func", "Time", "Reds", "Memory",
                "MsgQ", "Current Function"]}]}.

    We also want to get the table of those 20 processes having the largest
    message queues in each 5 seconds. We just one to list the processes.
    So, we open another stream.

        {stream, process_list, [{lines, 20}, {interval, 20}, {sort, msg_q}]}.

    The result tell us similar things as above. Observe that the tag is
    different.

        {ok,
         {etop, _StreamID = 2},
         [{data_structure, table},
          {headers, ["Pid", "Name or Initial Func", "Time", "Reds", "Memory",
                     "MsgQ", "Current Function"]}]}.

    Waiting for results..

    The first result of our first query arrives.

        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.

    After 5 seconds

        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.

    After another 5 seconds

        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.

    And another

        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.


    And the first result of the second query arrives.

        {{etop, _StreamID = 2},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.

            {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.

    Now, the second query is not needed anymore, so we close the stream.

        {stream, close, {etop, _StreamID = 2}}.

    Of course, the results of the first query still arrives.

        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.
        {{etop, _StreamID = 1},
         [{data,
           [[<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3],
            [<0.25.0>, group:server/3, '-', 5980, 55152, 0, group:server_loop/3]]}]}.



  - Observer. We want to get the table of those 10 processes consuming the
  most memory in each 5 seconds. We may want to further investigate any of the
  processes.

    So, we open a stream.

        {stream,
         process_list,
         [interactive, {lines, 10}, {interval, 5}, {sort, memory}]}.

    We receive a confirmation that the stream has been initiated, its data will
    be marked with the {observer, 1} tag. We got information about the table's
    properties and about the table's type -- interactive.

        {ok,
         {observer, _StreamID = 1},
         [{data_structure, table},
          {type, interactive},
          {headers = ["Pid", "Name or Initial Func", "Time", "Reds", "Memory",
                      "MsgQ", "Current Function"]}]}.

    And the first piece of data arrives.

        {{observer, _StreamID = 1},
         [{data, [
            [{<0.25.0>,
             {actions,
             [[{command, {command, kill_process, [{reason, kill}, {pid, <0.25.0>}]}},
               {label, "Kill process"}],
              [{command, {command, process_info, [{pid, <0.25.0> }]}},
               {label, "Process info"}],
              [{command, {command, process_info, [{key, messages}, {pid, <0.25.0> }]}},
               {label, "Process messages"}],
              [{command, {command, process_info, [{key, dictionary}, {pid, <0.25.0> }]}},
               {label, "Process dictionary"}],
              [{command, {command, process_state, [{pid, <0.25.0> }]}},
               {label, "Process state"}],
               ...]}},
              group:server/3,
              '-',
              5980,
              55152,
              0,
              group:server_loop/3],
            [...]]}]}

    With the process in each row we can do some further actions. These shall be
    presented in a local menu appearing after clicking on the process.
    We click on the menuitem kill process. As a result the same command will be
    sent that was defined as an action in the previous result.

        {command, kill_process, [{reason, kill}, {pid, <0.25.0>}]}.

    We don't expect to receive any result.

    We click on another menuitem. Command is sent.

        {command, process_info, [{key, dictionary}, {pid, <0.25.0> }]}

    The result describes the data structure, which is a table, and its data
    as well.

        {ok, [{data_structure, table}, {type, static}, {headers = ["Key", "Value"]},
              {data, [['$initial_call', {wxe_server, init, 1}], [ref, xyzerfe]]}]}.

    We close the table just rendered.

    Another piece of data from the stream still arrives.

        {{observer, _StreamID = 1},
         [{data, [
            [{<0.25.0>,
             {actions,
             [[{command, {command, kill_process, [{reason, kill}, {pid, <0.25.0>}]}},
               {label, "Kill process"}],
              [{command, {command, process_info, [{pid, <0.25.0> }]}},
               {label, "Process info"}],
              [{command, {command, process_info, [{key, messages}, {pid, <0.25.0> }]}},
               {label, "Process messages"}],
              [{command, {command, process_info, [{key, dictionary}, {pid, <0.25.0> }]}},
               {label, "Process dictionary"}],
              [{command, {command, process_state, [{pid, <0.25.0> }]}},
               {label, "Process state"}],
               ...]}},
              group:server/3,
              '-',
              5980,
              55152,
              0,
              group:server_loop/3],
            [...]]}]}


## Important observations

- process_list. It was served in two different ways. The difference was the
possibility of taking further actions. It is okay to have multiple features for
the same thing as the feature *best fitting* to the given arguments will be
selected using deterministic heuristics.

- Force GC and User commands. Both can be a long running task. Although
commands can be said to be synchronous calls, it cannot be implemented as is.
Consider that the evaluation of the call takes more time then the timeouts in
our system. We don't want to either loose the result of the command or make the
browser hanging. Therefore, once the command is arrived and accepted, the
corresponding plugin just makes a promise that it will send the result of the
command tagged by a  unique ID. But how will the result arrive? We will use
again streaming. The only difference is that this stream will only contain one
piece of data so after transferring data the stream auto closes itself.
Let's revisit the examples.

  - Force GC.

          {command, force_gc, []}.

    We receive either {error, Reason} or  {ok, {force_gc, _StreamID = 1}, []}.

    And after a while we receive

          {{{command, force_gc}, _StreamID = 1},
           [{data_structure, value},
            {label, "Force GC:"},
            {data, "ok"}].

  - Execute user commands.

        {command, eval, [{expr, "lists:sort([1,2,3]) + 42"}]}.

    We receive either {error, Reason} or  {ok, {eval, _StreamID = 1}, []}.

    And after a while we receive

          {{{command, eval}, _StreamID = 1},
          [{data_structure, value},
          {label, "Result"},
          {data,
              "{'EXIT',{badarith,[{erlang,'+',[[1,2,3],42],[]},
              {erl_eval,do_apply,6,[{file,"erl_eval.erl"},{line,661}]},
              {erl_eval,expr,5,[{file,"erl_eval.erl"},{line,434}]},
              {shell,exprs,7,[{file,"shell.erl"},{line,684}]},
              {shell,eval_exprs,7,[{file,"shell.erl"},{line,639}]},
              {shell,eval_loop,3,[{file,"shell.erl"},{line,624}]}]}}"}]}.

- Uniform responses. Defining multiple protocols without a real need brings
unnecessary complexity to the code, so it should be avoided. We saw that
commands are something synchronous, but some of them can only be realised as
a sequence of asynchronous responses. We use stream for long running commands,
but we still have synchronous responses for short running commands. So, the
infrastructure should only use streams. After a request arrives the
infrastructure shall reply with either {error, Reason} or {ok, StreamID,
StreamDescription}. Then the upcoming replies will be sent as {StreamID,
StreamData}. So, lets see three examples.

  - Kill the shell.

          {command, kill_the_shell, []}
          ----> {ok,
                  {{command, kill_the_shell}, 1},
                  [{data_structure, value}, {label, "Result"}, {stream, auto}].
          ----> {{{command, kill_the_shell}, 1}, "Shell got killed."}.

  - Force GC.

          {command, force_gc, []}
          ----> {ok,
                  {{command, force_gc}, 1},
                  [{data_structure, value}, {label, "Result"}, {stream, auto}].
          ----> {{{command, force_gc}, 1}, "GC was successfully forced."}.

  - Request an interactive process listing.

          {stream,
           process_list,
           [interactive, {lines, 10}, {interval, 5}, {sort, memory}]}.
          ----> {ok,
                {observer, 1},
                [{data_structure, table},
                 {type, interactive},
            {headers = ["Pid", "Name or Initial Func", "Time", "Reds", "Memory", "MsgQ", "Current      Function"]}]}.
          ---->
          {{observer, 1},
           [
             [{<0.25.0>,
              {actions,
              [[{command, {command, kill_process, [{reason, kill}, {pid, <0.25.0>}]}},
                {label, "Kill process"}],
               [{command, {command, process_info, [{pid, <0.25.0> }]}},
                {label, "Process info"}],
               [{command, {command, process_info, [{key, messages}, {pid, <0.25.0> }]}},
                {label, "Process messages"}],
               [{command, {command, process_info, [{key, dictionary}, {pid, <0.25.0> }]}},
                {label, "Process dictionary"}],
               [{command, {command, process_state, [{pid, <0.25.0> }]}},
                {label, "Process state"}],
                ...]}},
               group:server/3,
               '-',
               5980,
               55152,
               0,
               group:server_loop/3],
             [...]
           ]}
          {stream, close, {observer, 1}}.


- As the available plugins and the applications running on the managed node can
  only define the available streams and commands, we can only organise them to
  categories during development, whilst their exact set of arguments that they
  accept turns out during runtime.


- A GUI dynamically rendering data by interpreting it is essential.


- Not directly available commands. Have you realised the commands for killing
  a certain process in the stream coming from observer? These are further options,
  will be presented as menuitems in the local menu of a certain process.
  These commands are also announced, but may not be directly usable by the users.
  The reason is simple, just consider that you have an option to kill any
  process but you don't want to kill random pids. It doesn't make sense.


- Possibility to install receivers for streams. Features are very likely to be
  used not only by the users, but also by other Wombat functionalities. What is
  convenient for the GUI is not the perfect choice for other Erlang processes. So,
  the infrastructure should only communicate with other Erlang processes using
  asynchronous message passings. Streams will be manifested by a dedicated
  process and also wrappers for the communication will be provided as well.


- Sometimes no concurrency allowed. Some features require exclusive usage
  because the concurrent usage is either impossible or dangerous. Just consider
  that concurrent streams relying on the same tracer can't be active, whilst
  concurrently forcing GC is possible but may have a negative impact on the node.
  Features impossible to use concurrently are Observer, ETop. Managing
  configuration, forcing GC, and perhaps evaluating users' commands are dangerous.
  However, the other features can still be used concurrently.


- History of the operations carried out. By adding tooling into Wombat Wombat
  becomes an active agent that does have side-effect. Therefore, all operations
  carried out together with their results shall be recorded. Logging it as
  notifications provides a history that can be easily browsed by the users.


## Node family level

Until that point features were examined from the node's perspective. However,
commands including configuration changes make sense on node family
level.Performing a command on a node family means doing the same thing on all
the nodes. But, here what really makes sense is likely to be different regarding
to separate features. So, we define the layer only in general.

It is an abstraction layer that aggregates the same feature provided by the
plugins running on the managed nodes. But can a command be executed on a node
family level if not all its nodes are up? How can the results of the commands
executed on the nodes be aggregated? It cannot be determined in general, thus,
for each feature a server should be defined that contains the business logic and
coordinates the feature on each node. The services provided by the servers,
however, are similar to the base features their aggregate. Namely, these
services are not always available. So, if there is a change in their status, the
GUI should be notified and should work accordingly.

As we can have multiple node families, for each node family an instance of this
server should be started when all its preconditions are satisfied and of course
it shall be disabled or stopped whenever it is necessary. That results in a
supervisor tree having dynamically added children realising the same feature and
a global manager server dealing with the servers.

## High level design

- Once a plugin on a managed node determines what it can exactly do with its
node, it will announce each of the commands and streams as standalone
capabilities. The description of the capability contain

    - the category, and

    - the list of the arguments it accepts. The arguments shall have
      descriptions as well.

- Main capabilities' categories.

    - Configurator. Although the high-level functionality to change the
      configuration of an application is the same, applications vary on the
      persistence used for storing their configs. The most commonly used is using
      configuration files, however, some applications use different storage.
      For instance, Mongoose uses Mnesia tables to store values. Ergo, the details
      are domain specific knowledge, which may be handled by its plugin.

    - Explorer. Either static or periodically updated information about a
      certain process or about the VM. Has only a very small side-effect on the
      managed node, if it has any.

    - Executor. Real commands (force GC, evaluate user commands). Executed for
      their side-effects. May not have any meaningful result upon its status. A
      history about the executions shall be available for the users.

- At Wombat's side a standalone application - wo_services - is responsible for
  providing tooling. It transforms the capabilities reported by the plugins into
  services. Services then are published which can be used by both the GUI and
  other Wombat modules.


- Requests for services are identified by a unique ID.

- When a request for a service arrives the wo_services application handles it.

  1. generates a unique ID for the request

  2. registers the request

  3. checks its preconditions
  4. outsources the request

     1. selects the capability best / first fit to the request

     2. dispatches the request to the corresponding plugin implementing the
      capability

     3. waits for a quick response from this plugin

     4. if the response is positive, this plugin will be responsible for handling
      the request, otherwise go to step 4.1.

  6. updates its information about the request with the result

     1. monitors the plugin handling the request

     2. changes the status of the request (started, aborted, ..)

     3. stores additional information received from the plugin

  7. depending on the response it either

     - initialises a stream dedicated for the request and then makes a promise to
      the requester, or

     - returns an error.

- Features at Wombat's side. Once a Wombat process dealing with an exact node
family decides the features it wants to provide it registers its features in
the wo_services application. These features will also be services, available for
any users. Their deviation from the plugin's features is hidden from the users.
However, there is a main difference. These features control themselves. In order
to fulfil the user's request it can request for other requests, concurrently
receive data from different streams, and can directly write any data to the
stream dedicated for the user's request.


- Realisation of commands and streams. Initiating both shall be done by REST
calls. The response for the REST call contains either an error or a promise.
The promise contains the tag to identify the response and perhaps some
instructions for rendering. The GUI should subscribe to the stream using the tag
defined in the promise via websocket. (The first piece of data can be sent only
after the subscription is completed. So, data arrived before having any active
listeners shall be buffered.) Then the streams will be pushed to the GUI
(via websocket). Closing the streams can be initiated by the GUI. Streams
related to services causing no side-effects will be automatically closed if no
active subscribers present. That will cause terminating the data-source (the
plugin pushing data to Wombat) as well.


- Announcing capabilities. Once a plugin can realise a feature (or is now unable
to realise it), it shall announce it by publishing its actual complete list of
capabilities using wombat_plugin:announce_capabilities/1.


- Tooling capabilities. A tooling capability is a tuple complying with the type
wombat_types:capability/0. The type of the capability falls into configurator,
explorer or executor. Information about the concrete feature shall include

    - description: Textual description of the feature.
    - type: configurator | explorer | executor.
    - label: Textual representation of the label shall be presented by
    the GUI.
    - feature: The key to request the feature.

        - force_gc,
        - process_list,
        - soft_purge_old_modules,
        - kill_the_shell,
        - eval,
        - kill_process,
        - process_info,
        - process_state,
        - [config_mgmnt].

    - is_internal: Whether the feature should be directly available for the users.
    If it isn't then the GUI shouldn't be aware of it. (Consider the observation
    about the 'Not directly available commands.')
    - is_exclusive: Whether the feature can be concurrently used or not.
    - priority: The more preferred the implementation is, the larger the number
      is. Default is 0.
      (Consider that the same feature may be provided by multiple plugins.)
    - arguments: List of mandatory options. That is used by the wo_services
      application to match the proper service to a given request.
    - options: List of both mandatory and optional options.  (Interface of the
      feature)

          Option :: [{option_name, The key shall be used in the request},
                     {option_label, The label of the option shall be presented},
                     {option_type, Type of the value, can be any | number |  enum | list},
                     {option_default, any. The default value},
                     {option_enabled, true | false.
                      Whether the option can be seen and directly set by the users.},
                     {option_values,
                      [OptionValueBinStr
                       iff it is an enum},
                     {listitem_type, Option | [ Option1, Option2, OptionN ]
                       iff it is a list}]

## Requirements for the GUI

- To generate a page listing all available services belonging to the node or
the node family selected by the user.
- To generate the input form of the selected service.
- To post the generated input form to Wombat's REST API and handle the response.
Response contain either information both about the stream and about the
rendering or an error message.
- To subscribe to a stream via WS.
- To close streams not needed any more via WS.
- To display the list of active streams. Users can list them, switch between
them and close their own views into any of them.
- To render data based on the received instructions. Rendering information
is sent only once for efficiency reasons, but can be always queried while
the stream is active.
- To handle actions included by received data as interactive menuitems and
present their results.


## Overall architecture

The architecture is organised into 2 hosts: Wombat and the managed node.
Description mainly follows data and control flow.

### Glossary

- plugin feature. A realisation of a feature by a certain plugin on a certain
  managed node. Described as a capability.
- service. Available for users. A service is either
  - a node family level feature provided as a service by a process at Wombat
    side.
  - the result of merging and aggregating some capabilities announced by
    different plugins running on the same managed node.

  Information about a service provided to users:
  - its authority (node family or managed node),
  - its mandatory arguments, and
  - instructions to render its form.
- service request. A service request is the request to execute a service using
  the user given inputs sent by users. A service request will be transformed into
  either a list of plugin features or a node family feature.
- request. A request is the execution of a certain feature using the user given
  inputs. Each request provides its exclusive stream.
- stream. A stream is exclusively dedicated to a certain active request.
  A stream include as many bidirectional channels as many subscribers it has.
  The data and the control flow from Wombat to subscribers are identical on all
  channels. However, it doesn't hold in the reverse way. Control flows can be
  different on the channels of a certain stream.
- channel. A channel is between Wombat and a certain subscriber enabling the
  flow of data and control related to a stream.



### Wombat's side

#### wo_plugins application
Main entry and exit points.

##### wo_plugins_controller

It can be used as is, is the exit point used to request a concrete plugin.

##### wo_plugins_dispatcher

It is the entry point for control information and data both originated from the
plugins. It will be extended to accept and carefully dispatch the new messages
to the wo_services_dispatcher.

##### wo_services_dispatcher

It is a new dispatcher that receives plugin's stream data and the capabilities
announced.

Stream data (data, info, command) is dispatched to the
wo_services_request_stream process that is dedicated to the request to which the
stream data belongs. The corresponding API functions to be called are

    wo_services_request_stream:stream_data/2
    wo_services_request_stream:stream_info/2
    wo_services_request_stream:stream_command/2


Capabilities are dispatched to the wo_services_locator process. Only those
capabilities are important that are are plugin features. The corresponding API
function to be called is

    wo_services_service_locator:replace_plugin_implementations/4.

It monitors the following processes:

- wo_services_locator,
- mnesia_controller,
- mnesia_monitor,
- gproc.


#### wo_services application

It is a new application. It provides the services, accepts and execute service
requests, supports  streams. Owner of the ETS tables used by the manager and the
locator processes. The important types and environmental variables are listed
and then its processes are described.

During the start phase the application module

1. creates the following ETS tables:

   - wo_services_request_catalogue (refer to wo_services_request_mgr)
   - wo_services_request_info  (refer to wo_services_request_stream)
   - wo_services_internal_request_info  (refer to wo_services_request_stream)
   - wo_services_service_info  (refer to wo_services_service_locator)
   - wo_services_service_internal_info  (refer to wo_services_service_locator)

   The tables are created by calling the following API functions:

       wo_services_request_stream:init_tables/0
       wo_services_request_mgr:init_tables/0
       wo_services_service_locator:init_tables/0

2. adds dispatch rule to Cowboy to register its REST interface.



##### TYPES used in the wo_services application

Request args are the input given by the user, which are used to execute a
certain request. Value can be either

- a plain binary if the type of the option for which the value is given is a
  primitive type, or
- a list of lists containing key value pairs if the type of the option is list.

    -type request_args() ::
            list({KeyBinStr :: binary(), Value :: binary() | [request_args()]}).


display_info() describes for the GUI how to display streamed data. Created by
the plugin process.


    -record(display_info, {data_structure = value       :: value | table,
                           label = <<"">>               :: binary(),
                           table_headers = []           :: list(binary()),
                           is_interactive = false       :: boolean()}).

    -type display_info() :: #display_info{}.


Promise consists of the pid of the executor process, the display_info() and the
stream_info(). Created by the wombat_plugin behaviour as the response returned
to the wo_services_request_stream process.

    -record(promise, {display_info :: display_info(),
                      stream_info  :: stream_info(),
                      executor_pid :: pid()}).

    -type promise() :: #promise{}.


stream_info() describes for the GUI how to handle stream. Created by the
wombat_plugin behaviour using the #execution_info{} record returned by the
plugin process.

    -record(stream_info, {is_periodic = false      :: boolean(),
                          is_interruptible = false :: boolean()}.

    -type stream_info() :: #stream_info{}.



request_info() is the public view of a request. Created and handled by the
request_stream process. Stored in ETS and then in Mnesia. Used by the GUI.

    -record(request_info, {request_id :: binary(),
                           label :: binary(),
                           object_type :: node | family,
                           object :: binary(),
                           start_date :: erlang:timestamp(),
                           end_date :: erlang:timestamp(),
                           arguments :: list(argument_info()),
                           info_messages :: list(stream_info_message()),
                           display_info :: display_info(),
                           stream_info :: stream_info(),
                           data :: stream_data()}).

    - type request_info() :: #request_info{}.


stream_info_message() is the information already streamed to the GUI.
Updated by the request_stream process. Used by the GUI.

    -record(stream_info_message, {date :: erlang:timestamp(),
                                  message :: binary()}).

    - type stream_info_message() :: #stream_info_message{}.


argument_info() stands for a user input given when the service request was
submitted. It stores the label of the input to which the user entered Value.
Used by the GUI. Created by the request_stream process. Refer to request_args/0
to understand the second type of the type union ([[argument_info()]]).

    -type argument_info() :: {Label :: binary(), Value :: binary()} |
                             {Label :: binary(), [[argument_info()]]}.


internal_request_info() is the private view of a request. Used to map request
IDs to request_stream processes, to cleanup after these processes and to
maintain the list of subscribers. Updated by the request_stream process.

    -record(internal_request_info, {request_id :: binary(),
                                    pid :: pid(),
                                    executor_pid :: pid(),
                                    executor_controller :: {NodeId :: binary()
                                                            PluginName :: atom()} | pid(),
                                    subscribers :: sets:set(pid())}.

    -type internal_request_info() :: #internal_request_info{}.

feature_request() is a request for a feature, the body of the POST request sent
by the GUI to request an action defined in the data streamed. It consists of a
feature description and the user given input. It is later transformed into a
service_request().

    -record(feature_request, {feature :: feature_name(),
                              object_type :: node | family,
                              node :: node(),
                              request_args :: request_args()}.

    -type feature_request() :: #feature_request{}.

service_request() is the request for a specific service, the body of the POST
request sent by the GUI. It identifies the service requested and contains the
user given input.

    -record(service_request, {service :: service_id(),
                              request_args :: request_args()}.

    -type service_request() :: #service_request{}.

request() contains control information used by the wo_services_request_mgr
process. It is used to determine the public running requests, to ensure global
invariants, to support the recovery of the wo_services_request_mgr process.

     -record(request, {request_id :: binary(),
                       status :: checked | starting | running | closing,
                       request_stream_pid :: pid(),
                       label :: binary(),
                       is_internal :: boolean(),
                       service_id :: service_id(),
                       object_type :: node | family,
                       object :: binary()}).

service_id() uniquely identifies a service available for exactly one node or
node family under exactly one type (executor, configurator, explorer). A service
groups some implementations of a certain feature. It is an opaque type, only the
wo_services_service_locator can extract information from it.


     -opaque service_id() :: {ObjectType :: node | family,
                              Object :: binary(),
                              Type :: configurator | explorer | executor,
                              Feature :: feature_name(),
                              GroupID :: reference()}.

feature_name() is the name of the feature. Can have multiple implementations.
The set of implemented features are documented and known by the users and by
the plugin developers.

     -type feature_name() :: term().

service_info() is the public view of a certain service. It includes all
necessary properties to categorise, display, generate the input form and request
for a service. Used by the GUI, the request_stream, and the request_mgr.
Note that object_type, object, is_internal fields are hidden from the GUI.

     -record(service_info,{service_id :: service_id(),
                           object_type :: node | family,
                           object :: binary(),
                           is_internal :: boolean(),
                           label :: binary(),
                           description :: binary(),
                           options :: list(option())}).

     -type service_info() :: #service_info{}.

option() defines how the GUI should display, handle and add the result of a user
input to the request.

     -record(option,{name :: binary(),
                     label :: binary(),
                     type :: any | number | enum | list,
                     default :: binary(),
                     enabled :: boolean(),
                     mandatory :: boolean(),
                     values = [] :: [binary()],
                     listitem_type = [] :: [option()]}).

     -type option() :: #option{}.


Data to be streamed is typed as follows.

    -type stream_data() :: stream_data_value() | stream_data_table().

    -type stream_data_value() :: plain_value() | ineractive_value().

    -type plain_value() :: binary().

    -type interactive_value() :: #interactive_value{} | plain_value().
    -record(interactive_value, {data :: binary(),
                                actions :: list(action())}).

    -type action() :: #action{}.
    -record(action, {label :: binary(),
                     feature_request :: #feature_request{}}).


    -type stream_data_table() :: stream_data_plain_table() |
                             stream_data_interactive_table().

    -type stream_data_plain_table() :: list(list(plain_value())).

    -type stream_data_interactive_table() :: list(list(interactive_value())).



##### Envs

Maximum number of finished requests to be kept. Its default value is 20.

    {wo_services, [{request_history_limit, 20}]}.

Whether executing service requests that belong to the executor type is allowed.
Its default value is true.

    {wo_services, [{enable_executor, true}]}.

##### wo_services_sup

The main supervisor of the application. All other processes are its children.

Restart strategy: one_for_one, tolerates a medium number of restarts.

Its children, which are listed in the preferred start order:

1. wo_services_db

2. wo_services_event_sup

3. wo_services_service_locator

4. wo_services_request_stream_sup

5. wo_services_request_mgr


##### wo_services_lib

Library module of the functions frequently used by the wo_services application.
Types used by only the wo_services application are defined and exported by
this module.

- request_info/1 hides the complexity of retrieving the #request_info{} record
  from users. It first queries the record from wo_services_request_mgr. If the
  record cannot be found, it queries the wo_services_db module to gather the
  record.

    request_info(ReqID :: binary()) ->
      ReqInfo :: request_info() | {error, ReasonBinStr :: binary()}.

- requests/3 hides the complexity of retrieving request lists from users.
  If the running requests should be returned it calls the
  wo_services_request_mgr:running_requests/2 function, whilst if the finished
  requests are required it calls the wo_services_db:finished_requests/2
  function.

    requests(Type :: running | finished,
             ObjectType :: node | family,
             Object :: binary()) -> list({ReqID :: binary(), Label :: binary()).


##### wo_services_db

It is a global gen_server process that handles the history of the requests and
also limits the history's size. The size of the history is a user configuration
option, the corresponding env is

    {wo_services, [{request_history_limit, 20}]}.

It ensures the schema of the wo_requests_info Mnesia table:

- #request_info{} records,
- key is #request_info.request_id,
- type is set,
- disc copy table.

It provides the following API functions:

- finished_requests/2 is a library function not calling into the gen_server
  process. It collects the ReqID and the label of those requests that were
  working with the Object given. Dirty read is preferred.

      finished_requests(ObjectType :: node | family, Object :: binary()) ->
          list({ReqID :: binary(), Label :: binary()).

- request_info/1  is a library function not calling into the gen_server
  process. It first looks up for the request_info record of the request
  identified by ReqID. Dirty read is preferred.

      request_info(ReqID :: binary()) ->
        ReqInfo :: request_info() | {error, ReasonBinStr :: binary()}.

- save_request_info/1 is a non-blocking call into the server process. It inserts
  the ReqInfo given into the Mnesia table and may remove the oldest request not
  to violate the history's limit. By calling into the gen_server process
  we can serialise concurrent writes and can ensure the history's limit.

      save_request_info(ReqInfo :: request_info()) -> ok.

##### wo_services_service_locator

It is a locally registered gen_server.

- It uses two ETS tables:

  - wo_services_service_info,
  - wo_services_service_internal_info.

- It maintains the service catalogue containing the currently available services
  related to any nodes and any node families. The service catalogue contains the
  public information about each of the services (refer to service_info()).

  The service catalogue is stored in a public ETS table. The server provides
  API, which is a lib function, to query the public parts of the catalogue:

      service_info(Type :: configurator | explorer | executor,
                   ObjectType :: node | family,
                   Object :: binary()) ->
                       list(service_info()).

      service_info(service_id()) -> service_info() | {error, not_found}.

- It hides executor services, if the value of the enable_executor env is false.

- It keeps track of the implementations provided by the plugins for features
  on the nodes.

  The implementations provided by the plugins are registered by the
  wo_services_dispatcher using the following API function.

      replace_plugin_implementations(ObjectType :: node | family,
                                     Object :: binary(),
                                     PluginName :: atom(),
                                     Capabilities :: wombat_types:capabilities()) ->
          ok.

  It transforms the plugin features announced by different plugins running on
  the same managed node into services.

- It allows processes at Wombat side to register the implementations of node and
  node family level features. These implementations are handled as the
  implementations provided by plugins, they will be transformed to services.
  These implementations are registered in its catalogue together with
  information about how to initiate requests for these services. That is the
  only difference, namely, how the corresponding wo_services_request_stream
  process should interact with the process implementing the feature.

  Note that replace_plugin_implementations/4 is a non-blocking call whilst
  register_implementation/4 is a synchronous call.

       register_implementation(ObjectType :: node | family,
                               Object :: binary(),
                               ExecutorPid :: pid(),
                               Capability :: wombat_types:capability()) ->
           {ok, ServiceID :: service_id()} | {error, Reason :: term()}.

- It allows processes at Wombat's side to withdraw services using the following
  API functions:

        unregister_implementation(ObjectType :: node | family,
                                  Object :: binary(),
                                  ExecutorPid :: pid(),
                                  CapabilityID :: wombat_types:capability_id()) ->
            ok | {error, not_found}.

        unregister_service(ServiceID :: service_id()) ->
            ok | {error, not_found}.

- It derives services from the registered implementation. The transformation is
  done by grouping the features together based on both their types and their
  tooling_features and then joining their options together. The results of the
  transformation are the services that are registered in its catalogue together
  with the implementations used to derive each of the services. Note that the
  transformation is a non-injective surjective function.

  Transformation is done whenever any implementation is removed or a new
  implementation is added. To understand how the transformation works, study
  the axioms and and assume the following

  - before the transformation begins we have a set of services for which the
    axioms are hold,

  - transformation aborts whenever its result would violate the axioms,

  - the result of a transformation, which is the new set of services, satisfy
    the axioms.

  Axioms

  1. Each service is either a primitive service or a composite service.

     A service is a primitive service if and only if it has exactly one
     implementation.

     A service is a composite service if and only if it has more
     virtual implementations.

  2. The intersection of the implementation sets of any two different primitive
     services is empty.

  3. For all implementations of all composite services exactly one primitive
     service exists the implementation of which equals to a definite
     implementation of a composite service.

  4. The intersection of the implementation sets of any two different composite
     services is empty.

  5. For any two different composite services (s1, s2) which provide either only
     exclusive or only non-exclusive and either only internal or only
     non-internal implementations for the same feature on the same object the
     followings hold:

     1. The intersection of the two arguments sets belonging to an
        implementation of s1 and to an implementation of s2 is empty.

     2. The intersection of the two arguments sets belonging to s1 and s2 is
        empty.

  6. For all implementations included by the same service the followings hold:

     - Implements the same feature.
     - Dedicated to the same object.
     - Have the same type (configurator, executor, explorer).
     - Both are exclusive or both are non-exclusive.
     - Both are internal or both are non-internal.
     - The intersection of their arguments are non-empty or both argument sets
       are empty.
     - The options defined by both are generalisable.

  7. The properties of any composite service is derived by the properties of
     its implementations.

     - Implements the same feature.
     - Dedicated to the same object.
     - Have the same type (configurator, executor, explorer).
     - Either exclusive or non-exclusive.
     - Either internal or non-internal.
     - Its arguments are those that are also the arguments of all its
       implementation. Its argument set equals to the intersection set of all
       the argument sets defined by its implementation.
     - Each of its options is defined as an option at least by one
       implementation. The options defined by two or more implementations are
       generalised.


  Now, study how generalising options work. You may refer to [the Types
  supported by wo_services](https://gist.github.com/viktoriafordos/b754bcd50062045d5695#file-supported_types-md) in advance.

  1. Two options must be generalised if and only if they have the same names.

  2. Two options can be generalised if and only if

     1. the type of both is either ANY or NUM, or

     2. one's type is the subtype of the other's type (NUM < ANY, ENUM < ANY), or

     3. the type of both is ENUM, or

     4. the type of both is LIST and the components of both lists can be
        generalised.

  3. The properties of the option generalising two ANY or two NUM options are

      - preserves the type,

      - is mandatory if and only if both options were mandatory,

      - is enabled if and only if any of the options was enabled,

      - have default value if and only if any of the options had a default value.

  4. The properties of the option generalising two ENUM options are

      - preserves the type,

      - is mandatory if and only if both options were mandatory,

      - is enabled if and only if any of the options was enabled,

      - the set of its possible values is the union of both original sets of
        values,

      - have default value if and only if any of the options had a default value.

   5. The properties of the option generalising a NUM and ANY or an ENUM and
       ANY options are

      - its type is ANY,

      - is mandatory if and only if both options were mandatory,

      - is enabled if and only if any of the options was enabled,

      - have default value if and only if any of the options had a default value.

   6. The properties of the option generalising LIST options are

      - its type is LIST,

      - is mandatory if and only if both options were mandatory,

      - is enabled if and only if any of the options was enabled,

      - its components are all generalised by the same generalising rules.


- It listens to node and node family events and also monitors the processes at
  Wombat's side to keep the catalogue up-to-date.
  If a node is down or a node family is deleted or a monitored process died then
  it removes the services the source of which matches to the subject of the event.

- It raises events whenever its catalogue has been changed. The listener to the
  events it raises is the wo_services_service_event_summariser server.

- It resolves service requests. The set of implementations belonging to the
  service requested is ordered. The order of the implementations are the basis of resolving service requests.

  We say an implementation (imp_1) is more preferred than another (imp_2) iff

  1. imp_1 has more mandatory arguments than imp_2.

  2. imp_1 has as many mandatory arguments as imp_2 has AND
     the priority of imp_1 is not less than the priority of imp_2.

            imp_1 precedes imp_2 <=>
            size(imp_1.args) > size(imp_2.args) OR
               (size(imp_1.args) == size(imp_2.args) AND
                  imp_1.priority >= imp_2.priority)

  The ordered set is the input of the resolving algorithm. The implementations
  having mandatory arguments that are not given by the service request are
  dropped from this set.

  It can return:

    - A list of implementations for a certain feature ordered using heuristics.
      The best fitting implementation is the first item of the list.
      Each implementation can be given by either a plugin or a process.

    - An error if it couldn't resolve the request.


           resolve(ServiceRequest :: service_request()) ->
               [{wo_topo:node_id_bin(),
                 atom(),
                 wombat_types:capability_id()} |
                {wo_topo:family_id_bin(),
                 pid(),
                 wombat_types:capability_id()} |
                {wo_topo:node_id_bin(),
                 pid(),
                 wombat_types:capability_id()}] |
               {error, ReasonBinStr :: binary()}.


- It transforms feature requests to service requests. Feature requests are
  defined by either an implementation as part of the stream data, or by a
  Wombat's side process. It is used in cases when the service cannot be exactly
  specified by its service ID. Feature requests must specify directly the object
  type, the node, the feature and the arguments.

  The transformation begins with the determination of the object. It can be
  exactly derived from the node name and from the object type. Then those
  services are selected that provide implementations for the given feature on
  the determined object. These services are further examined to remove those
  services that cannot recognise any of the request arguments or define such
  mandatory arguments that are not included by the request arguments. The
  primitive service having the most preferred implementation will be the result
  of the transformation. (The remaining services are ordered as described at the
  resolving algorithm.)

        feature_request2service_request(FeatureRequest :: feature_request()) ->
            ServiceRequest :: service_request() |
            {error, ReasonBinStr :: binary()}.


- It provides an API function to check whether the execution of a new service
  request would violate global invariants. The decision is made using the
  following rules.

  1. If nothing is running (RunningServices is an empty list) then the new
     request is allowed.

  2. If the new request is non-exclusive then it is allowed.

  3. New exclusive primitive service request is allowed if and only if none of
     the running requests uses an implementation that is either the same as
     the primitive service has or included by the implementation set of the
     composite service which is derived using the primitive service.

  4. New exclusive composite service request is allowed if and only if none of
     the running requests uses an implementation that is included by the
     implementation set of the composite service.

  If {error, Reason :: binary()} is returned the request should be denied and
  the error should be returned to the user.

       is_allowed(NewService :: service_id(),
                  RunningServices :: list(service_id())) -> boolean() |
                  {error, Reason :: binary()}.


- Failure handling. We expect that the failure of the locator doesn't effect the
  service catalogue. Therefore, whenever the server restarts it should be able
  to repair its catalogue to be totally consistent.

##### wo_services_request_mgr

It is a global gen_server.

It maintains the public, named ETS table called wo_services_request_catalogue.
It reads from wo_services_request_info ETS table. It provides data to be stored
in Mnesia.

- It pre-processes feature requests provided by the wo_services_rest_handler
  module. It transforms the feature request to a service request using the
  wo_services_service_locator module. If the transformation succeeded, it
  start pre-processing the service request returned, otherwise, returns an
  error.

      new_feature_request(FeatureRequest :: feature_request()) ->
          {ok, ReqID :: binary(), Label :: binary()} |
          {error, ReasonBinStr :: binary()}.

- It pre-processes service requests provided by the wo_services_rest_handler
  module. Pre-processing always includes some checks and perhaps the
  initialisation of executing the service requested. The API to initiate the
  processing of a service request is

      new_request(ServiceRequest :: service_request()) ->
          {ok, ReqID :: binary(), Label :: binary()} |
          {error, ReasonBinStr :: binary()}.

  Steps of processing service requests.

    1. Generates a unique request ID for each service request.

    2. Checks whether the execution of the service request just arrived
       would violate the global invariants (i.e.: exclusivity constraint) using
       wo_services_service_locator module.
       If so, the new request will be rejected and

           {error, ReasonBinStr :: binary()}

       will be returned. Otherwise the processing continues.

    3. Saves the new request together with all information into its request
       catalogue. The status of the request is checked.

    4. Starts a wo_services_request_stream process for each service request.
       The RequestID and the the service request is used to initiate the new
       process.
       To start the new process the wo_services_request_stream_sup supervisor's
       API function will be called.

    5. Monitors the new process.

    6. Updates the information about the request with the new process in its
       request catalogue. The status of the request is changed to starting.

    7. Exit with a noreply from the handle_call callback. The reply will be sent
       to the client when either the promise or the error arrives from the
       wo_services_request_stream process.

    8. Waits for the wo_services_request_stream process to either

       - make a promise using the promise/1 API function, or

       - to terminate with the reason of the failure.

    8. Updates its request catalogue. The status of the request is now running.

    9. Finally, it returns either

         - the ReqID, or
         - the error to the wo_services_rest_handler.

            {ok, ReqID :: binary(), Label :: binary()} |
            {error, ReasonBinStr :: binary()}.

       If an error is returned it removes the request from the request catalogue.

- It pre-processes custom service requests initiated by special processes at
  Wombat's side, not directly by the users. The processing allows these
  processes to override the control properties of the request and to
  provide custom check to ensure the global invariants. The corresponding API is

      new_custom_request(Request :: service_request() | feature_request(),
                         Props :: [{is_internal, boolean()}],
                         TestFunc :: default | fun((service_id())-> boolean())) ->
          {ok, ReqID :: binary()} | {error, ReasonBinStr :: binary()}.

  By overriding some control properties, processes can make their
  requests internal, for instance. Internal requests are hidden from the users -
  they are not included in the result of the REST calls - in the request lists.
  TestFunc will be sequentially evaluated on each of the requests if and only if
  it is a one-arity function. If its result is true for all the requests then
  the execution can be started otherwise the request should be denied to avoid
  violating global invariants.

- It maintains its request catalogue. The catalogue is stored in a public, named
  ETS table, called wo_services_request_catalogue. The following information
  about each request is kept in its request catalogue:

        -record(request, {request_id :: binary(),
                          status :: checked | starting | running | closing,
                          request_stream_pid :: pid(),
                          label :: binary(),
                          is_internal :: boolean(),
                          service_id :: service_id(),
                          object_type :: node | family,
                          object :: binary()}).

- It provides API function, which is a lib function not calling into the server
  process, to query the running requests.

      running_requests(ObjectType :: node | family, Object :: binary()) ->
          list({ReqID :: binary(), Label :: binary()).

  The result contains only those requests for which all the following
  constraints hold:

    - is_internal = false,

    - status = running.

  For each running request its ReqID and its label are returned.

- It provides an API function, which is a lib function not calling into the
  server process, to query request info about a given request. Actually, it
  calls wo_services_request_stream:get_request_info/1 to gather the record.

      request_info(ReqID :: binary()) ->
          wo_services_lib:request_info() | {error, not_found}.

- It provides an API function, which is a lib function not calling into the
  server process, to map request IDs to wo_services_request_stream processes.

      request_stream_pid(ReqID :: binary()) ->
          {ok, RequestStreamPid :: pid()} | {error, ReasonBinStr :: binary()}.

  The function serves to purposes:

    - to dispatch stream data coming from the wo_services_dispatchers or from
      the executor processes at Wombat's side to the wo_services_request_stream
      process.

    - to route new subscribers coming from the WS to the
      wo_services_request_stream process.

- It raises events whenever a request's status transitions into running or
  closing states or its catalogue has been changed. The listener for its events
  is the wo_services_request_event_summariser server.

- It monitors its wo_services_request_stream processes. If any of these
  processes terminates normally then it

    1. updates the status of the request to closing in its request catalogue,

    2. persists the #request_info{} record of the request by the wo_services_db
       process,

    3. removes the request from its catalogue,

    4. raises event about its catalogue.

- If a wo_services_request_stream process terminates abnormally then it

    1. updates the status of the request to closing in its request catalogue,

    2. cleans up after the request by calling
       wo_services_request_stream:cleanup/1 function,

    3. persists the #request_info{} record of the request by the wo_services_db
       process,

    4. removes the request from its catalogue,

    5. raises event about its catalogue.

- It provides API function, which is a gen_server:cast, for the
  wo_services_request_stream processes allowing the processes to make promises.

      promise(ReqID :: binary()) -> ok.

- Failure handling. We expect that the failure of the manager doesn't effect the
  active requests. Therefore, whenever the server restarts it should be able to
  repair its catalogue to be totally consistent with the active requests.


##### wo_services_request_stream_sup

Supervisor of wo_services_request_stream processes.

- one_for_one,
- temporary children.

It provides API functions to start and stop its children.

    start_request_stream(ReqID :: binary(), ServiceRequest :: service_request()) ->
        {ok, RequestStreamPid :: pid()} | {error, Reason :: any()}.

    stop_request_stream(ReqID :: binary()) ->
        ok | {error, not_found}.

It provides API function to find its child using ReqID.

    get_request_stream(ReqID :: binary()) ->
        {ok, RequestStreamPid :: pid()} | {error, not_found}.


##### wo_services_request_stream (per request)

It is a gen_server dedicated for a service request. It is responsible for
executing a service request given as argument and maintaining channels for
subscribers.

It creates, maintains and updates

- the ReqInfo :: request_info() of its request in the public, named ETS table
  called wo_services_request_info.

- the InternalReqInfo :: internal_request_info() of its request in the public,
  named ETS table called wo_services_internal_request_info.


- Initialising. The ReqID and the ServiceRequest are the arguments of its init/1
  function. It stores these values in its state. Creates and then inserts its
  #internal_request_info{} record to the ETS table.

      #internal_request_info{request_id = ReqID,
                             pid = self(),
                             executor_pid = undefined,
                             executor = undefined,
                             subscribers sets:new()}.

     Finally, it enters resolving state.


- Its logical states (should not appear in the implementation):

  - resolving. It is the initial state. It uses wo_services_service_locator to
    resolve service requests. Once it has the #service_info{} record and its
    preference list containing the possible executors of the given service
    request, it enters the outsourcing state. If it cannot resolve the service
    request, it terminates with specific reason enabling the
    wo_services_request_mgr to reply to the client.

  - outsourcing. It sequentially goes through the preference list and try to
    outsource the request. It continues outsourcing the request while the
    executors respond that the request is out of their scope. It stops, if a
    promise is made to fulfil the request or an error returned from an executor
    that considers itself capable for the task.
    In the former case, it

      1. starts monitoring the executor process defined by the promise,

      2. creates and then inserts its #request_info{} record to the ETS table

              Arguments = create_args(ServiceRequest, ServiceInfo).

              #request_info{request_id = ReqID,
                            label = ServiceInfo#service_info.label,
                            object_type = ServiceInfo#service_info.object_type,
                            object = ServiceInfo#service_info.object,
                            start_date = os:timestamp(),
                            end_date = undefined,
                            arguments = Arguments,
                            info_messages = [],
                            display_info = Promise#promise.display_info,
                            stream_info = Promise#promise.stream_info,
                            data = undefined}.

      3. updates its #internal_request_info{} record in the ETS table

              OldInternalReqInfo#internal_request_info{
                  executor_pid = Promise#promise.executor_pid,
                  executor_controller = ExecutorController}.

              ExecutorController :: ExecutorServerPid :: pid() |
                                    {NodeId :: wo_topo:node_id_bin(),
                                     PluginName :: atom()}.

         ExecutorController can be used to request the executor to stop
         (refer to cleanup/1). That happens if a) user requests to stop the
         execution or b) the request_stream process has abnormally terminated
         and the request_mgr process cleans up after the process.

      4. makes a promise to the wo_services_request_mgr (refer to promise/1),

      5. enters streaming state.

    Otherwise, if none of the possible executors has accepted the request then
    it deletes its #internal_request_info{} record from the ETS table and
    terminates with specific reason enabling the wo_services_request_mgr to
    reply to the client.

  - streaming. In this state it maintains channels for subscribers

      - it is a repeater from the executor's perspective (i.e.: forwards stream
        data to subscribers), and
      - it is a forwarder from the subscriber's perspective (i.e.: listens to
        stream commands sent by subscribers).

    It is its final state. It leaves this state when it receives the

        {stream_command, ReqID :: binary(), close}

    message from its executor pid. That means the executor has completed the
    request, won't send any further message, therefore, the stream can be
    closed. The wo_services_request_stream process forwards the stream command
    to its subscribers and terminates with reason normal.

- Requests can be outsourced to the possible executors that are contained by the
  preference list.

        [{NodeID, PluginName, CapabilityID} |
         {NodeID, ExecutorPid, CapabilityID} |
         {NodeFamilyID, ExecutorPid, CapabilityID}].

  Requests can be outsourced either to plugins (ex. Mnesia) or to Wombat's side
  processes (ex. node family configurator server). Plugins are identified by

        {NodeID :: binary(),
         PluginName :: atom(),
         CapabilityID :: wombat_types:capability_id()}.

  Outsourcing the request to a certain plugin is initiated by calling

        wo_plugins_controller:plugin_call(NodeID,
                                         PluginName,
                                         Request).

  Where Request is

        {may_start_request,
         ReqID :: binary(),
         CapabilityID :: wombat_types:capability_id(),
         ReqArgs :: request_args()}.

  The result of the call can be

        {ok, Promise :: promise()} |
        {out_of_scope, ReasonBinStr :: binary()},
        {error, ReasonBinStr :: binary()}.

  Wombat's side processes are identified by

        {NodeOrNodeFamilyID :: binary(),
         ServerPid :: pid(),
         CapabilityID :: wombat_types:capability_id()}.

  Outsourcing the request to a Wombat process is done by message passings.
  RequestMsg is sent to ServerPid and then the wo_services_request_stream
  process waits until the Reply arrives or a timeout occur. Timeouts are handled
  in the same way if {error, ReasonBinStr} had been returned.


        RequestMsg =
            {{may_start_request,
              ReqID :: binary(),
              CapabilityID :: wombat_types:capability_id(),
              ReqArgs :: request_args()},
             RequestStreamPid :: pid()}.

        Reply =
            {ReqID :: binary(), {ok, Promise :: promise()}} |
            {ReqID :: binary(), {out_of_scope, ReasonBinStr :: binary()}},
            {ReqID :: binary(), {error, ReasonBinStr :: binary()}}.

- Subscribers can open a channel by calling the following API function. It
  should be a non-blocking call that adds the SubscriberPid to the set of
  its subscribers (both in its state and in the #internal_request_info{} record).

      subscribe(RequestStreamPid :: pid(), SubscriberPid :: pid()) -> ok.

      subscribe(ReqID :: binary()) -> ok | {error, Reason :: any()}.


- Subscribers can send a few stream commands.
  A subscriber can close (unsubscribe) its channel, which is its own view into
  the stream, or terminate the whole stream (stop).

  When the process unsubscribes from the stream the wo_services_request_stream
  process removes SubscriberPid from the set of its subscribers.

  Terminating streams means stopping the execution of the request that will
  affect all active views into the stream. However, terminating streams are
  not always possible. If the stream is not interruptible then the command will
  be simply ignored by the executor. To send stream commands, the following
  non-blocking API functions are provided.

      unsubscribe(RequestStreamPid :: pid(), SubscriberPid :: pid()) -> ok.

      unsubscribe(ReqID :: binary()) -> ok | {error, Reason :: any()}.

      stop(RequestStreamPid :: pid()) -> ok.

      stop(ReqID :: binary()) -> ok | {error, Reason :: any()}.


- Stream data can be forwarded to subscribers by calling the following API
  function. It should be a non-blocking call.

      stream_data(RequestStreamPid :: pid(), StreamData :: stream_data()) -> ok.

      stream_data(ReqID :: binary(), StreamData :: stream_data()) ->
          ok | {error, Reason :: any()}.

  The process will

    1. update the #request_info.data value with the StreamData in the ETS table,

    2. send the following message to all its subscribers.

           {stream_data, ReqID :: binary(), StreamData :: stream_data()}.


- Stream info can be forwarded to subscribers by calling the following API
  function. It should be a non-blocking call.

      stream_info(RequestStreamPid :: pid(), InfoBinStr :: binary()) -> ok.

      stream_info(ReqID :: binary(), InfoBinStr :: binary()) ->
          ok | {error, Reason :: any()}.

  The process will

    1. create the stream_info_message

           InfoMessage = #stream_info_message{date = os:timestamp(),
                                              message = InfoBinStr}.

    2. insert to the beginning of the list stored in #request_info.info_messages

    3. send the following message to all its subscribers.

           {stream_info, ReqID :: binary(), InfoMessage}.


- Stream commands can be forwarded to subscribers by calling the following API
  function. It should be a non-blocking call.

      stream_command(RequestStreamPid :: pid(), close) -> ok.

      stream_command(ReqID :: binary(), close) -> ok | {error, Reason :: any()}.

  The process will

    1. update #request_info.end_date with the current timestamp,

    2. send the following message to all its subscribers,

           {stream_command, ReqID :: binary(), close}.

    3. delete its #internal_request_info{} record from the ETS table,

    2. terminate with reason normal.

- It monitors the executor process. If the
  monitored process died then it notifies its subscribers and terminates.
  In this case subscribers receive 2 messages. The first message is information
  about the termination, which is also stored in the #request_info.info_messages,
  whilst the second message is a stream command to close
  channels as no further data related to this request will be received.

      {stream_info, ReqID :: binary(), InfoMessage :: stream_info_message()}.

      {stream_command, ReqID :: binary(), close}.

- It requests the executor to stop executing the request if it has no
  subscriber for a while and the request is an interruptible request, otherwise
  even with no receivers it waits until the executor finishes and closes the
  stream and terminates.

- Requesting the executor to stop begins with creating an info message about
  the reason of the termination (requested by the user, or the stream seems to
  be useless, or the request_stream process crashed).

  Requesting the executor to stop is different in case of a plugin is the
  executor and in case of a Wombat's side process is the executor.

  Requesting a plugin to stop is done by calling

       wo_plugins_controller:plugin_cast(NodeID,
                                        PluginName,
                                        {may_stop_request, ReqID :: binary()}).

  Requesting a Wombat's side process to stop is done by sending

       {may_stop_request, ReqID :: binary()}

  to its ExecutorPid.

- Failure handling. The wo_services_request_stream process doesn't handle
  its failure, it just simply terminates. Instead, the wo_services_request_mgr
  cleans up after any unexpected, abnormal termination of the
  wo_services_request_stream process by calling cleanup/1.

      cleanup(ReqID :: binary()) -> ok.

  As this process is the bridge between the subscribers and the executor, both
  party need to be notified. So the cleanup/1 function

    1. update #request_info.end_date with the current timestamp,

    2. notifies the executor just like the wo_services_request_stream process
       does when it request its executor, which is either a plugin process
       or a process at Wombat's side, to stop. It uses the
       #internal_request_info.executor to determine the executor.

    3. notifies the subscribers by sending 2 messages. The first message is
       information about the termination, whilst the second message is a stream
       command to close channels as no further data related to this request will
       be received.

           {stream_info, ReqID :: binary(), InfoBinStr :: binary()}.

           {stream_command, ReqID :: binary(), close}.

    4. deletes the #internal_request_info{} record from the ETS table.

- It provides an API function, which is a lib function not calling into the
  server process, to query request info about a given request.

      get_request_info(ReqID :: binary()) ->
          wo_services_lib:request_info() | {error, not_found}.



##### wo_services_event_sup

Supervisor of the event summariser servers.

##### wo_services_service_event_summariser

It summarises events raised by the wo_services_service_locator to decrease the
load on the wo_services_service_locator. Very similar to the server summarising
the metric meta data events.

##### wo_services_request_event_summariser.

It summarises events raised by the wo_services_request_mgr to decrease the load
on the wo_services_request_mgr. Very similar to the server summarising the
metric meta data events.

##### processes at Wombat's side

Node family level configurator. To be specified, not the part of the framework.
(It uses the framework).

#### wo_rest application

- Define the REST interface.

wo_services_rest_handler, which is a new REST handler, will provide the REST API.
The REST API should allow users to

- GET the available services per node and per node family,
- GET the list of active requests, to which users can subscribe,
- GET the list of finished requests, to which users can open a view,
- GET the request info about a given request enabling the GUI to render data,
- POST a service request.

##### GET the available services per node and per node family

- GET all configurator services available for a node family

      GET /api/service/service-info/configurator/node-family/NODEFAMILYID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

- GET all configurator services available for a node

      GET /api/service/service-info/configurator/node/NODEID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

- GET all explorer services available for a node family

      GET /api/service/service-info/explorer/node-family/NODEFAMILYID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

- GET all explorer services available for a node

      GET /api/service/service-info/explorer/node/NODEID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

- GET all executor services available for a node family

      GET /api/service/service-info/executor/node-family/NODEFAMILYID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

- GET all executor services available for a node

      GET /api/service/service-info/executor/node/NODEID

  Action

        wo_services_service_locator:service_info/3

  Response

      [service_info()].

##### POST a request

- POST a service request

       POST /api/service_request


       {id: ServiceID,
        arguments: [ {OptionName, OptionValue} ]}.

  Action

        wo_services_request_mgr:new_request/1

  Response

      {error_text: ErrorText} |
      {request_id: ReqID,
       request_title: ReqTitle}.

- POST a feature request defined as an action in data streamed

       POST /api/feature_request

       {object_type: node | node-family,
        object: NodeName | NodeFamilyID,
        feature: OpaqueFeature,
        arguments: OpaqueArguments}

  Action

        wo_services_request_mgr:new_feature_request/1

  Response

      {error_text: ErrorText} |
      {request_id: ReqID,
       request_title: ReqTitle}.


##### GET information about requests

- GET the request info about a request

       GET /api/request/request-info/ReqID

  Action

        wo_services_lib:request_info/1

  Response

      {error: ErrorText} | request_info().

- GET all running requests per node

       GET /api/request/node/NodeID

  Action

        wo_services_lib:requests/2

  Response

       [{request_id: ReqID,
         request_title: ReqTitle}].

- GET all running requests per node family

       GET /api/request/node-family/NodeFamilyID

  Action

        wo_services_lib:requests/2

  Response

       [{request_id: ReqID,
         request_title: ReqTitle}].

- GET all finished requests per node

       GET /api/request/history/node/NodeID

  Action

        wo_services_lib:requests/2

  Response

       [{request_id: ReqID,
         request_title: ReqTitle}].

- GET all finished requests per node family

       GET /api/request/history/node-family/NodeFamilyID

  Action

        wo_services_lib:requests/2

  Response

       [{request_id: ReqID,
         request_title: ReqTitle}].


#### wo_core application (WS)

The websocket process will be extended to

  - Subscribe to events about changes related to both services or requests by
    calling

        wo_services_service_event_summariser:subscribe_service_catalogue_changed().
        wo_services_request_event_summariser:subscribe_request_catalogue_changed().

  - Receive and then send messages to the client.

    - Forwarding stream info messages. The websocket process receives

            {stream_info, ReqID :: binary(), InfoBinStr :: binary()}.

        that is converted as below and then sent to the client

             [{<<"type">>, <<"streamInfo">>},
              {<<"requestId">>, ReqID},
              {<<"message">>, InfoBinStr}].

    - Forwarding stream commands. The websocket process receives

            {stream_command, ReqID :: binary(), close}.

        that is converted as below and then sent to the client

            [{<<"type">>, <<"streamCommand">>},
             {<<"requestId">>, ReqID},
             {<<"command">>, <<"close">>}].

    - Forwarding stream data. The websocket process receives

            {stream_data, ReqID :: binary(), StreamData :: stream_data()}.

        that is converted as below and then sent to the client

             [{<<"type">>, <<"streamData">>},
              {<<"requestId">>, ReqID},
              {<<"data">>, to_json(StreamData)}].

        On the value of #action.feature_request term_to_binary should be called first.

    - Forwarding events about changes related to either services or requests.
       The websocket process receives

             #wo_event{event = {service_catalogue_changed | request_catalogue_changed,
                                {node | node-family, UUIDBin}}}.

       and converts them using the corresponding event converter module to

             [{<<"type">>, <<"serviceCatalogueChanged">>},
              {<<"objectType">>, <<"node">> | <<"node-family">>},
              {<<"id">>, UUIDBin}].

             [{<<"type">>, <<"requestCatalogueChanged">>},
              {<<"objectType">>, <<"node">> | <<"node-family">>},
              {<<"id">>, UUIDBin}].

        that are sent to the client.

  - Interpret stream related commands (subscribe, unsubscribe and stop) and then
    dispatch the commands to the corresponding process.

    - Subscribing to streams. The following message should be sent by the client
      to subscribe to the stream of the request identified by ReqID.

          <<"requests.stream.subscribe:", ReqID/binary>>

      The websocket process should call

          wo_services_request_stream:subscribe(ReqID)

      to add itself as a subscriber for the stream.

    - Unsubscribing from streams. The following message should be sent by the
      client to unsubscribe from the stream of the request identified by ReqID.

          <<"requests.stream.unsubscribe:", ReqID/binary>>

      The websocket process should call

          wo_services_request_stream:unsubscribe(ReqID)

      to unsubscribe itself from the stream.


    - Stopping streams. The following message should be sent by the client to
      stop the request identified by ReqID.

          <<"requests.stream.stop:", ReqID/binary>>

      The websocket process should call

          wo_services_request_stream:stop(ReqID).



### Managed node's side

Most service requests will be completed by plugin processes running on the
managed nodes. The following brief overview sums up the functionality.

1. Boundary Layer -- wombat_plugin_controller and wombat_plugin_relay.
   Main entry and exit points for requests. wombat_plugin_controller allows
   calls and casts into the plugin processes. Resolves the pid of the plugin
   based on its name. wombat_plugin_relay forwards plugins' messages to
   wo_plugins_dispatcher. Both can be used as is.


2. Services Layer -- wombat_plugin
   General part of the behaviour realised by the plugins.

     - It ensures constraints on the interface.

       - Pre- and also post-processes requests: cleans input by removing
         information unnecessary for the realisations of the plugins, tags the
         result with information required for Wombat.

       - Validates and wraps control information and data both originated from
         the realisations of the plugins.

     - It manages capabilities.

     - It provides a framework for the concrete realisations of the features.

       - Defines callbacks so the plugin developers only need to implement the
         specific parts.

       - Supervises and controls ongoing requests.

       - Handles failures: either by recovering from them or by reporting the
         total failure to Wombat.

       - Executes requests (user implemented callbacks + the general parts of
         the execution).


3. Feature Layer -- plugins (plugin modules)
   It implements the wombat_plugin behaviour. It can build only on the
   wombat_plugin's API (API functions of the wombat_plugin and the
   wombat_plugin_utils modules). It realises features, which are announced
   as capabilities, by implementing the callbacks.

#### Types defined in wombat_plugin and wo_plugins applications

Request args are the input given by the user, which are used to execute a
certain request.

    -type request_args() ::
            list({KeyBinStr :: binary(), Value :: binary() | [request_args()]}).

display_info() describes for the GUI how to display streamed data. Created by
the plugin process.


    -record(display_info, {data_structure = value       :: value | table,
                           label = <<"">>               :: binary(),
                           table_headers = []           :: list(binary()),
                           is_interactive = false       :: boolean()}).

    -type display_info() :: #display_info{}.

execution_info() describes for the wombat_plugin behaviour how to execute the
request. Created by the plugin process.

    -record(execution_info, {period = once        :: once | non_neg_integer(),
                             retry_after = never  :: never | non_neg_integer(),
                             max_retries = 0      :: non_neg_integer()}.

    -type execution_info() :: #execution_info{}.

Promise consists of the pid of the executor process, the display_info() and the
stream_info(). Created by the wombat_plugin behaviour as the response returned
to the wo_services_request_stream process.

    -record(promise, {display_info :: display_info(),
                      stream_info  :: stream_info(),
                      executor_pid :: pid()}).

    -type promise() :: #promise{}.

stream_info() describes for the GUI how to handle stream. Created by the
wombat_plugin behaviour using the #execution_info{} record returned by the
plugin process.

    -record(stream_info, {is_periodic = false      :: boolean(),
                          is_interruptible = false :: boolean()}.

    -type stream_info() :: #stream_info{}.

Data to be streamed is defined in Section wo_services application.

#### Wombat_plugin behaviour

*TODO (after the first version is stable): optionally accumulate the data
streamed*

It allows plugin processes to

- announce capabilities,
- perform requests,
- stream results.

It allows Wombat to

- start requests,
- stop requests.

- It manages capabilities. After the plugin process's init the behaviour calls
  the capabilities/1 callback to retrieve capabilities. Plugin processes are
  also allowed to change their capabilities by calling the
  wombat_plugin:announce_capabilities/1 API during runtime. Capabilities are
  always checked and then pushed to Wombat by the general part of the plugin.
  The checks ensure that only correct capabilities will be pushed to Wombat.
  Considering tooling capabilities the constraints are

    - each capability of a certain plugin has unique capability ID,
    - type of the capability is valid,
    - both the label and the feature of the capability is defined,
    - the options of the capability have unique names,
    - the type of the options are valid,
    - option for each argument is defined in the options of the capability.

- It provides handlers to interact with the wo_services_request_stream process.
  2 kinds of commands can arrive: to start executing a request, and to stop the
  execution of a running request.

  - Initiate the execution of a request with the given user inputs is a
    synchronous call into the plugin's process. Note that the real execution is
    not done here!

        handle_call({may_start_request,
                     ReqID :: binary(),
                     CapabilityID :: wombat_types:capability_id(),
                     ReqArgs :: request_args()},
                     _From :: reference(),
                     State :: #state{}) ->

                {ok, Promise :: promise()} |
                {out_of_scope, ReasonBinStr :: binary()},
                {error, ReasonBinStr :: binary()}.

    Steps of the execution:

    1. Initiates the request using the implementation of the init_request/4
       callback. The callback can make 3 decisions:

         - Serving the request is out of its scope. For instance, consider a
           special configurator plugin that changes only the configs of the
           Mongoose application. If it is asked to change the config of the
           riak_core application, it is simply not capable of performing the
           change. In this case it returns

                   {out_of_scope,
                    ReasonBinStr :: binary(),
                    NewPluginState :: wombat_types:plugin_state()}

         - Serving the request is its authority but the provided arguments are
           incorrect. I.e. in theory, to perform the request it has all
           necessary inputs, but the validation of the inputs have been failed.
           For instance, consider the observer plugin that receives the
           <<"ETC">> binary as the value of its interval argument for which it
           only accepts binaries can be converted to non-negative integers. In
           this case it returns

                   {error,
                    ReasonBinStr :: binary(),
                    NewPluginState :: wombat_types:plugin_state()}

         - Is able and willing to serve the request. Serving the request is
           its authority and also all mandatory arguments are given, have been
           checked and considered to be valid.
           In this case it initialise the execution of the request and makes a
           promise that will be returned

                   {ok,
                    DisplayInfo :: display_info(),
                    ExecutionInfo :: execution_info(),
                    NewPluginState :: wombat_types:plugin_state()}.


       If the callback's response is negative (either error or out_of_scope)
       then the execution ends and the response returned by the callback is
       returned as the result of the call.
       Otherwise the plugin accepted the request and specified all information
       necessary both to display its upcoming data and to handle its stream
       (DisplayInfo, ExecutionInfo).

    2. Records some information about the request (ReqID, CapabilityID,
       ExecutionInfo) in its state.

    3. Schedules the execution of the request immediately after the current call
       is completed by sending {exec_req, ReqID} to itself.

    4. First, it produces the stream_info record:

         - is_periodic is true if and only if execution_info.period is not once.
         - is_interruptible is true if and only if execution_info.period is not
           once.

       And then creates the promise record (Promise):

         - display_info is the #display_info{} returned by the plugin process.
         - stream_info is the #stream_info{} record just produced.
         - executor_pid is the pid of the plugin process (self()).

    5. Returns {ok, Promise :: promise()} as the result of the call.


  - Terminating the request is a non-blocking call into the plugin process,
    which is not guaranteed to succeed.

        handle_cast({may_stop_request, ReqID :: binary()}, State :: #state{}) ->
            ok.

    Checks whether the execution of the request is interruptible using
    information stored in its state about requests. If so, it finalises the
    request.

- Finalising requests. When the execution has been finished (either normally or
  abnormally) or the user has requested to terminate the stream the plugin
  process initiates optional cleanup by calling the implementation of the
  cleanup_request/3 callback and then reports that

    - the execution has been finished with some reason by pushing

          {stream_info, ReqID :: binary(), InfoBinStr :: binary()}.

    - and also the stream shall be closed by pushing

          {stream_command, ReqID :: binary(), close}.

  Finally, it removes all information about the request from its state. That
  prevents the plugin process to perform the upcoming executions already
  scheduled.

- The execution of the requests are achieved by repeatedly evaluating it.
  The current evaluation is started by the plugin process when it processes the
  {exec_req, ReqID} message in its handle_info callback. It first checks whether
  the request is still valid (i.e.: it hasn't removed from its state). If so and
  the execution of the request is a short enough then it is evaluated as follows.

     1. Retrieves the stream data to be pushed to Wombat by calling the
         implementation of the execute_request/4 callback inside of a try-catch
         expression.

     2. May report the stream data towards Wombat by sending the following
         message.

             {stream_data, ReqID :: binary(), StreamData :: any()}.

     3. The result of successfully calling the execute_request/4 callback also
        defines whether the execution has been completed or it shall be
        continued. If the execution has been completed then the plugin process
        finalises the request, otherwise the plugin process resets the number of
        previous retries and then schedules the next evaluation of the request.

        If a runtime error occurred, the plugin process does the followings.

        1. Reports a pre-defined text about the unexpected failure as stream
           info towards Wombat by sending the following message.

                {stream_info, ReqID :: binary(), ErrorDesc :: binary()}.

        2. Finalises the request.

        If the result is an error, the plugin process does the followings.

        1. Reports the returned reason as stream info towards Wombat by sending
           the following message.

                {stream_info, ReqID :: binary(), ErrorDesc :: binary()}.

        2. Decides whether to schedule the next evaluation or to finalise
           the request by taking into account the defined retry strategy and
           the number of previous retries.

     4. Updates the plugin's state in its state.


  In cases when the execution would block the plugin process for a relatively
  long period of time, the implementation should spawn_link a worker process
  having a valid From to execute the request
  and then return {reply_later, State} as the result of calling
  execute_request/4. Then the plugin process will wait until the reply arrives
  or the spawned worker process terminates abnormally, and only after that it
  will stream data to Wombat. The worker process should be able to send its
  reply by calling

      wombat_plugin_services:request_reply(
        From :: wombat_plugin_services:from_ref(),
        Reply :: wombat_plugin_services:async_reply()).



#### Callbacks may be implemented by the plugin processes

- Implementing the init_request/4 callback is mandatory for all plugins that
  implements tooling features and therefore announce tooling capabilities.
  The goal is to decide whether to make a promise for handling the request.

  The callback should first determine that the request is covered by the
  authority of the plugin process or it is out of its scope. If the request is
  out of scope then

      {out_of_scope,
       ReasonBinStr :: binary(),
       NewPluginState :: wombat_types:plugin_state()}

  should be returned. Otherwise, the given user inputs should be validated. If
  any of the inputs are incorrect then

      {error,
       ReasonBinStr :: binary(),
       NewPluginState :: wombat_types:plugin_state()}

  should be returned. Otherwise, the plugin process is able and willing to serve
  the request. Serving the request is its authority and also all mandatory
  arguments are given, have been checked and considered to be valid. So, it
  initialises the request and provides information about the data pushed to the
  stream and how to handle the stream in response.

      {ok,
       DisplayInfo    :: display_info(),
       ExecutionInfo  :: execution_info(),
       NewPluginState :: wombat_types:plugin_state()}.

  Specification of the callback is the following.

         init_request(ReqID :: binary(),
                     CapabilityID :: wombat_types:capability_id(),
                     ReqArgs :: request_args(),
                     PluginState :: wombat_types:plugin_state()) ->

                     {ok,
                      DisplayInfo    :: display_info(),
                      ExecutionInfo  :: execution_info(),
                      NewPluginState :: wombat_types:plugin_state()} |
                     {out_of_scope,
                      ReasonBinStr :: binary(),
                      NewPluginState :: wombat_types:plugin_state()} |
                     {error,
                      ReasonBinStr :: binary(),
                      NewPluginState :: wombat_types:plugin_state()}.


    - DisplayInfo describes how to display stream data.

        - The data_structure defines how to display data will have been pushed.
          It can be value (a simple text) or table.  It is optional, its default
          value is value.

        - The label defines a text shall be displayed as label when the data is
          displayed. It is optional, its default value is the empty label.

        - The table_headers defines the headers of the table to be displayed. It
          makes only sense when table is the data_structure. It is optional, its
          default value defines a zero column table.

        - The is_interactive defines whether further actions will be present
          together with data. If so, the actions should be available for the
          users. It is optional, its default value is false.

    - ExecutionInfo describes how to execute the request.

        - The period describes how often data will be streamed.
          It can be once or a non negative number. Once means that data will be
          streamed only once and users are not allowed to stop the execution of
          the requests. Periodical requests can always be stopped by the users.
          The period of executing such requests is defined by the value of this
          option, namely, the given value defines the number of milliseconds
          elapsed between streaming data twice. It is optional, and its default
          value is once.

        - retry_after defines how the failures should be handled. It can be
          never or a non negative integer. Never means the evaluation of the
          request should be never retried, whilst the given number defines the
          number of milliseconds after the evaluation can be retried. Its
          default value is never.

        - retries_count defines the maximum number of attempts to evaluate
          the request in a row. If the number of attempts reaches the defined
          maximum, the plugin process gives it up and finalises the request. Its
          default value is 0, which means the plugin process will never retry
          the evaluation and gives up immediately after the first failure occurs.

- Implementing the cleanup_request/3 callback is optional, but it is
  recommended for plugins providing services built on the top of other services.
  In the callback the resources allocated for the request can be released, or
  services running can be stopped. The plugin process is prepared for the
  situation when the callback is not implemented. Thus, the plugin process won't
  crash. will be able to move forward, but no special cleanup will occur.

        cleanup_request(ReqID :: binary(),
                        CapabilityID :: wombat_types:capability_id(),
                        PluginState :: wombat_types:plugin_state()) ->

             {ok, NewPluginState :: wombat_types:plugin_state()}.


- Implementing the execute_request/4 callback is mandatory for all plugins
  that implements tooling features and therefore announce tooling capabilities.
  The goal is to really execute the request, to provide data to be streamed and
  then to be displayed based on the previously given DisplayInfo, and to define
  what will happen to the stream (should be closed or kept opened to continue
  the execution).

        execute_request(ReqID :: binary(),
                        CapabilityID :: wombat_types:capability_id(),
                        From :: wombat_plugin_services:from_ref(),
                        PluginState :: wombat_types:plugin_state()) ->

             {continue | close,
              no_data | {data, StreamData :: any()},
              NewPluginState  :: wombat_types:plugin_state()} |
             {reply_later, NewPluginState  :: wombat_types:plugin_state()} |
             {error,
              ReasonBinStr :: binary(),
              NewPluginState  :: wombat_types:plugin_state()}.


#### Important notes for the plugin developers

- Requests are executed in the plugin process. That means while the callbacks
  are being evaluated the plugin process cannot handle other tasks (i.e.: cannot
  push metrics, logs, alarms).

- Only the execution of non-periodic requests, the stream_period property of
  which is set to once, cannot be stopped by the users.

- The execution of periodic requests can always be stopped by the users. It is
  stopped by finalising the request instead of scheduling its next execution.
  Request being executed are not effected by stop commands, they are allowed to
  normally terminate.

- Information provided in the capabilities is used by

    - Wombat to create services by aggregating the capabilities that describe
      different implementations of the same feature.

    - Wombat to categorise the services. Services will be displayed under their
      category group on the GUI.

- Information provided in the display info (DisplayInfo) is used by

    - the wombat_plugin behaviour to control the execution of requests.

    - the GUI to display data to be produced and then streamed by the plugins.

    - the GUI to decide whether users are allowed to stop requests.


#### wombat_plugin_utils

Refer to wombat_plugin_services to constructors. I.e.:

- create_executor_capability/2,
- create_explorer_capability/2,
- create_configurator_capability/2,

- create_any_option/4,
- create_enum_option/5,
- create_list_option/5,

Optional API functions

- create_request_info/4
- create_execution_info/3.


## Dataflow diagram

[The diagram](https://gist.github.com/viktoriafordos/91dd3ac36345ef25be3f#file-dataflow-jpeg)
illustrates the data flow.

Remarks for nodes:

- WPD: wo_plugins_dispatcher of a node. (It simplifies the flow by joining
  dispatchers together. )

- Node: managed node.

- WPC: wo_plugins_controller of a node.

- L: wo_services_service_locator.

- ESum: wo_services_service_event_summariser.

- RS: wo_services_request_stream.

- RM: wo_services_request_manager.

- RSum: wo_services_request_event_summariser.

- WS: Websocket handler.

- ERL: Erlang processes at Wombat's side providing services.

- REST: wo_services_rest_handler.


Remarks for edges:

- Blueish edges: service related data.

- Purplish edges: request related data.

- Greenish edges: stream related data.

- Brownish edges: event related data.


## GUI

- new main menu called Services
- 4 new sub menus under "Services": Configure, Explore, Manage, Requests.

### Configure, Explore, Manage pages

- See [scratch of the pages](https://gist.github.com/viktoriafordos/91dd3ac36345ef25be3f#file-gui_services-jpeg).

- Add the same source listing that is used under Alarms page.

- Based on the selected node or node family and the page (any of configure,
  explore, manage) GET the list of available services via the REST API.

      GET /api/service/service-info/TYPE/OBJECTTYPE/OBJECTID

- Display a list view. Each service is a panel. The title of the panel is the
  value of the title property. All panels are closed by default. Users
  can collapse a panel by clicking on it. In this case the input form of the
  service appears in the body of the panel.

- Display the input form of the service. Use the properties (description,
  options) returned by the REST API.

- Actions triggered by clicking on the "Start request" button.

    1. Check whether all mandatory options are filled.

    2. Perform a POST request to /api/request. The body of the request contains

         - service_id: the ServiceID,
         - arguments: the list of user input. Each user input is described by
           {option.name, Value}.

    3. If the response is {error, ErrorText} then stay on the page, and display
       the ErrorText.

    4. If the response is {request_id, ReqID} then

       1. Subscribe to the stream of the new request using the value of the
          request_id property contained by the response.

       2. Add ReqID to the list of requests stored in the browser's local
          storage.

       3. Mark ReqID as the active view.

       4. Redirect the browser to the Requests sub menu.

- Actions triggered by clicking on the "Start request in background" button.

    1. Check whether all mandatory options are filled.

    2. Perform a POST request to /api/request. The body of the request contains

         - service_id: the ServiceID,
         - arguments: the list of user input. Each user input is described by
           {option.name, Value}.

    3. If the response is {error, ErrorText} then stay on the page, and display
       the ErrorText.

    4. If the response is {request_id, ReqID} then stay on the page, and display
       "Request started".

- Listen to service change events. If an event has been raised while the user
  stays on these pages then the list view should be refreshed.

- Refresh information stored about the available services when a new WS
  connection has been initiated.

### Requests page

- See

    - [scratch of the page](https://gist.github.com/viktoriafordos/91dd3ac36345ef25be3f#file-gui_requests-jpeg)
    - [scratch of the output panel](https://gist.github.com/viktoriafordos/91dd3ac36345ef25be3f#file-gui_output-jpeg)

- Add the same source listing that is used under Collected Metrics page.

- Based on the selected node or node family GET either

    - the list of running requests via the REST API

          GET /api/request/OBJECTTYPE/OBJECTID

    - the list of finished requests via the REST API

          GET /api/request/history/OBJECTTYPE/OBJECTID

- Listen to request change events. If an event has been raised while the user
  stays on the page then the request list should be refreshed.

- Refresh request list when a new WS connection has been initiated.

- New view into a request is opened by either

    - as the result of starting request, or

    - changing the checkbox of the request to be checked.

- Both opening a new view into a request and clicking on the tab of an already
  opened view activate the view.

- Multiple views can be opened at the same time. Each ReqID stored in the
  browser's local storage has an opened view.

- Only one view can be active at the same time. If there are some opened views
  while the user stays on the Requests page then one view is always active.

- The GUI subscribes only to the stream of the request belonging to the active
  view.

- Whenever the active view is being rendered or the WS connection was split and
  then established the GUI should

    1. temporary store the response (ReqInfo) of the

            GET /api/request/request-info/ReqID.

    2. initiate the detailed view based on ReqInfo

    3. subscribe to the stream of the request via WS (if the request is a
       running request)

    4. listen to WS (if the request is a running request)

       - If stream info arrives then it should be inserted into the Messages
         table as the first row.

       - If stream data arrives then the content of the Output panel should be
         redrawn with the fresh data.

       - If stream command (close) arrives then the GUI should display that the
         request has been finished.

- If the active view is changed or the user leaves the page then the GUI should

    1. unsubscribe from the stream of the request that was previously active via
       WS

    2. remove all temporary data (ReqInfo) stored about the request.

- When an opened view is closed by the user the reqID should be removed from the
  list of requests stored in its local storage.

- Views can be closed by either removing the check from the checkbox of the
  request or closing the tab of the view.

- If the request is interruptible, then the "Stop request" button should be
  visible.

- GUI should write the stream command (close) to the WS when the user clicks on
  the "Stop request" button.


### Local storage


Data must be permanently stored

- list of ReqIDs to which an opened view belongs,
- the ReqID of the active view.

Data must be temporary stored

- ReqInfo of the active view.

- List of either running or finished requests being displayed under the category
  belonging to the selected node or node family.

- List of ServiceInfos displayed as panels, which can be explorers,
  configurators or executors and are provided for a certain node or node family.

Data must not be permanently stored

- list of ServiceInfos,

- lists of running and finished requests,

- ServiceInfo,

- ReqInfo.

## The configurator service

### Real life experience
While I was doing Riak support many of my investigations resulted in changing
configs. There the usual approach was to change the config only on one node and
observe how it behaves. If the observed behaviour met with our expectation, we
recommended to change the config on the other nodes.

The lessons learnt

- Even if we want to change a global config it is good to first check its effect
only on one node.
- Happened to me many times that the customers either forgot to persist the new
config value or set it on the other nodes.
- Rolling restart is always preferred. That avoids temporal outages.

Conclusions

- Allow users to change a config locally (only on one node).
- Changing a global config only on one node is always an anomaly. Alarms help
users to remember they shall perform further actions.
- Rolling restart should be supported.

### Managing configs on node level
Although the high-level functionality to change the configuration of an
application is the same, applications vary on  storing their configs. The most
commonly used is  configuration files, however, some applications use different
storage. For instance, Mongoose uses Mnesia tables to store values. Ergo, the
details are domain specific knowledge, which may be handled by its plugin.
Nonetheless, a general plugin can be written as well.

The configurator plugins running on the nodes are super simple. They shall
provide the following services:

- get_configs. It shall return all the envs together with their values set for
the application. In regards of the general plugin it is a wrapper of
application:get_all_env(Application).
- {get_config, Key}. It shall return all the env which is defined by the Key
together with its value set for the application. In regards of the
general plugin it uses the application:get_env(Application, Par) to build the
result.
- {set_config, [{Key, Val, IsPersistent}]}. It shall set the values of the given
envs at once. If IsPersistent is true, the changes shall appear in the config
file on the managed node. In regards of the general plugin it may rely on
application:set_env/3 or application:set_env/4. It should be further
investigated.
- {reload_application, Application}. It reloads the application, since changing
the envs may only have effect after reload. In regards of the general plugin it
may rely on application:reload/1.

Services (capabilities) announced.

-  get_configs.
    {get_configs_general,
     [{description, "Retrieve an Erlang application's envs."},
      {type, configurator},
      {tooling_label, "Display Application's envs."},
      {tooling_feature, get_configs},
      {arguments, [application]},
      {options,
       [[{option_name, application},
         {option_label, "Application"},
         {option_type, enum},
         {option_default, application1},
         {option_values,
          [[{key, application1}, {label, "Application1"}],
           [{key, application2}, {label, "Application2"}]]}]]}]}.

- {get_config, Key}.
    {get_configs_general,
     [{description, "Retrieve an env of an Erlang application."},
      {type, configurator},
      {tooling_label, "Display Erlang application's env."},
      {tooling_feature, get_configs},
      {arguments, [application, env]},
      {options,
       [[{option_name, application},
         {option_label, "Application"},
         {option_default, "mnesia"},
         {option_type, any}],
        [{option_name, env},
         {option_label, "The key of the env"},
         {option_type, any}]]}]}.

- {set_config, [{Key, Val, IsPersistent}]}.
    {set_config_general,
     [{description, "Set an Erlang Application's env."},
      {type, configurator},
      {tooling_label, "Set an Erlang Application's env."},
      {tooling_feature, set_config},
      {arguments, [application, subject]},
      {options,
       [[{option_name, application},
         {option_label, "Application"},
         {option_type, any}],
        [{option_name, subject},
         {option_label, "List of env.s to be changed"},
         {option_type, list},
         {listitem_type,
          [[{option_name, env},
            {option_label, "The key of the env"},
            {option_type, any}],
           [{option_name, value},
            {option_label, "The value of the env to be set"},
            {option_type, any}],
           [{option_name, persist},
            {option_label, "Shall the change be persistent?"},
            {option_type, enum},
            {option_default, false},
            {option_values,
             [[{key, true}, {label, "Yes"}],
              [{key, false}, {label, "No"}]]}]
            ]}]]
        }]}.

- {reload_application, Application}.
    {reload_application_general,
     [{description, "Reload an Erlang application."},
      {type, configurator},
      {tooling_label, "Reload an Erlang application."},
      {tooling_feature, reload_application},
      {arguments, [application]},
      {options,
       [[{option_name, application},
         {option_label, "Application"},
         {option_type, enum},
         {option_default, application1},
         {option_values,
          [[{key, application1}, {label, "Application1"}],
           [{key, application2}, {label, "Application2"}]]}]]}]}.

#### Node family level

Until that point features were examined from the node's perspective. However,
commands including configuration changes make sense on node family
level. Performing a command on a node family means doing the same thing on all
the nodes. But, here what really makes sense is likely to be different regarding
to separate features. So, we define the layer only in general.

It is an abstraction layer that aggregates the same feature provided by the
plugins running on the managed nodes. But can a command be executed on a node
family level if not all its nodes are up? How can the results of the commands
executed on the nodes be aggregated? It cannot be determined in general, thus,
for each feature a server should be defined that contains the business logic and
coordinates the feature on each node. The services provided by the servers,
however, are similar to the base features their aggregate. Namely, these
services are not always available. So, if there is a change in their status, the
GUI should be notified and should work accordingly.

As we can have multiple node families, for each node family an instance of this
server should be started when all its preconditions are satisfied and of course
it shall be disabled or stopped whenever it is necessary. That results in a
supervisor tree having dynamically added children realising the same feature and
a global manager server dealing with the servers.

Node Family Configurator is dedicated to the global configurations of a certain
application loaded to the nodes belonging to a certain node family. Facts:

- Envs which are global configurations shall share the same value on all nodes.
- If the values are differ, then something went wrong, it is an anomaly.
- The values of a global configuration can only be changed iff the change
effects all the nodes to keep the consistency of the system.
- Global configurations are the subset of the local configurations. That means
changing a global configuration only on a certain node (i.e. locally ) is still
allowed. However, if it happens, it is still an anomaly. (It is sometimes useful
to try the new value only on one node before changing the value globally.)

A Node Family Configurator server is parametrised by the node family and by the
application. These are given as options during init.

The preconditions of the server are

- all the nodes are UP,
- the necessary services are available on all the nodes.

The server maintains the set of the global configurations' keys and their
values. This set is empty by default. An env become a global configuration iff
the user requests it.

The server periodically scans for anomalies. An alarm shall be raised iff
different values are set to any global configuration or a global configuration
is missing from any node of the family. The alarm shall be cleared iff the
values of all the global configurations are the same and also set on all the
nodes.

The server coordinates changing the value of a global configuration on all the
nodes. Changes are only allowed iff all the nodes are UP.

Changing the value of a global configuration is a process. While it is in
progress there is an anomaly. Some of the nodes have already been updated and
are using the new value, whilst the others are still using the old value. The
anomaly can be tolerated but only temporary until the end of the process. If
the process managed to change the value on all the nodes, the anomaly will
disappear, otherwise the anomaly remains. That is perfect, as the user will be
aware of the fact and can handle the situation.

A server shall provide the following services:

- {set_globals, NewSetOfGlobals}. Set the set of the global configuration keys.
- list_globals. Return the list containing all global configurations -- the envs
with their values.
- {set_config, [{Key, Val, IsPersistent}]}. Set the values of the given envs on
all the nodes. If IsPersistent is true, the changes shall appear in the config
files on the managed nodes.
- {reload_application, Application}. Reload the application (to use the new
envs) on all the nodes.


#### Centralised approach

This approach will be the part of ESL's contribution to SYNCFREE. This
will be the classic centralised approach. The Node Family Configurator processes
will be running on the Wombats and use Mnesia tables to store the global
configurations. Local configurations aren't stored, instead periodically queried
from the plguins. To successfully change the value of a global configuration all
the managed nodes must be reachable by the Wombats.

**Init**

1. A Riak cluster with 5 nodes has been added to Wombat. They form a new node
family.
2. The plugin architecture starts on each node reporting the capabilities.
3. Once all the configurators specified to riak_core is available and all the
nodes are UP, the process realising the server of the Node Family Configurator
for riak_core and the node family has been started. The set of global
configurations is empty in case of a new node family. If Wombat or the process
got restarted the set may not be empty.
4. The process checks the consistency of all global configs and maintains its
alarms.
5. The process announces the services.


**Set globals**

1. The user on the GUI requests the set of the global configuration to be
{max_concurrency}.
2. The process changes the set of the global configuration to be
{{max_concurrency, '$uninit'}}.

**Changing the set of global configurations or changes in the node family or
timeouts trigger the following actions.**

1. The process clears all of its active alarms.
2. The process queries the value of the max_concurrency env from all the nodes.
If the returned values are the same (e.g. it is 2), then saves the value to its
set {{max_concurrency, 2}}, otherwise
 - add all the different values to the set {{max_concurrency, [2, undefined]}},
 and,
 - raises an alarm.

**List global configurations**

1. The user requests the list of the global configurations. The server returns
its set.

**Change a global config**

1. The user requests changing a value of a global configuration.
Non-existing global configurations cannot be changed.
2. The process sequentially sends synchronuos commands to each node to change
the configuration.
3. Any failure stops the whole process, is reported to the user and raises an
alarm. (Alarm is raised only if there is a node running with the new value.)
Otherwise the below steps follow.
4. The new value is set to the key in the server's set.
5. Any alarm related to the changed configuration is cleared.
6. The success is reported to the user.

**Reload the application**

1. The user requests reloading the application on all the nodes.
2. The process concurrently sends synchronuos commands to each node to reload
the application.
3. The result is reported to the user.


**Examining the approach**

- relies on a centralised data store
- requires that all the managed nodes are reachable by the Wombats
- cannot handle network partitions

#### Decentralised approach

This approach will be the part of ESL's contribution to SYNCFREE. This
will be the decentralised approach. The Node Family Configurator processes
will be running on the Wombats and use AntiDote nodes to propagate changes and
commands.

The server here provides the same services as described previously, but use
AntiDote nodes to

- store configuration related data,
- propagate commands to the plugins running on the managed nodes.

For each managed node (and therefore for the plugins running on the managed node)
an AntiDote node is dedicated. The plugin knows how to communicate with its
AntiDote node.


3 bucket types will be created on the AntiDote nodes.
The local_configuration and the global_configuration bucket types capture the
current, real state of the system. Ergo, the first two are eventually consistent
with the current values of the envs on the managed nodes.
In contrary, the command bucket type represent the commands to be executed. It
indirectly describes the future state of the system.

Some further details on these bucket types.

- local_configuration. It will be a set storing the nodes' configurations
related to a certain application of a node family. Each item of the set will be
{Key, Node, Value}.
It defines a post-commit hook for each bucket that triggers the recalculation of
the values of the global configurations, if the modified configuration is also
a global configuration.
- global_configuration. It will be a map storing the global configurations
related to a certain application of a node family. The keys of the map will be
the keys of the configurations. The value of each key will be a set containing
the possible values set to the configuration on all the nodes.
Each bucket will has a pre-commit and a post-commit hook.
The pre-commit hook initialises each of the values of the global configurations
by deriving a set containing the different values of a global configuration from
the corresponding local_configuration bucket.
The post commit hook checks the derived set and notifies Wombat whether there is
an anomaly.
- command. Global set containing tasks to be executed.  To be performed on the
specified managed node. It is either 
 {set_config, Node, {Application, Key, Val}, IsPersistent, new}
or 
 {reload_application, Node,  Application, new}.
Post-commit hook notifies Wombat about the the new status.

Let's see an example. We have 3 node families, each of which use 20 applications.
Therefore we will have 3\*20 instances of the global_configuration bucket and
3\*20 instances of the global_configuration bucket. But, we will only have one
command bucket, as it is independent from  both the number of node families and
the number of applications.

The plugin has a different behaviour. It is more active compared to the
centralised approach.

- During init it PUT its envs to its local_configuration bucket.
- It periodically checks its envs. If any of the envs has changed, it PUT
its envs to its local_configuration bucket.
- It periodically GET the content of the command bucket. If the command is new
to the plugin then it performs the actions defined by the command.
If the command is {reload_application, Node,  Application, new} then it reloads
the specified application. After that it updates the command object with the
result and puts the new object to the command bucket.
If the command is {set_config, Node, {Application, Key, Val}, IsPersistent, new}
then the plugin changes the value of the given configuration (and persists it).
After that it PUT its envs to its local_configuration bucket.
Finally, it updates the command object with the result and puts the new object
to the command bucket.

**Init**

1. A Riak cluster with 5 nodes has been added to Wombat. They form a new node
family.
2. An AntiDote cluster with 5 nodes are started, join to the existing cluster
and its nodes are linked to the managed nodes. Note that the existing cluster
already contains the command bucket, as it a global single.
3. The plugin architecture starts on each node reporting the capabilities.
4. Once all the configurators specified to riak_core is available and all the
nodes are UP, the process realising the server of the Node Family Configurator
for riak_core and the node family has been started.
5. In case of a new node family Wombat creates 2 buckets in AntiDote for its
plugins, otherwise the buckets already exist.
6. The configurator plugins PUT the envs of the application running on their
managed nodes to their local_configuration bucket. Each put triggers the
recalculation of the map belonging to the global_configuration bucket. The
recalculation ends with notifications pushed to Wombat.
7. The process announces the services.


**Set globals**

1. The user on the GUI requests the set of the global configuration to be
{max_concurrency}.
2. The process PUT the new map corresponding to the new set of global
configurations to the global_configuration bucket. All values are '$uninit'.
As the values are '$uninit', the pre-commit hook determines the values and
post-commit hook notifies Wombat about the existence of the anomalies.

**Changing the set of global configurations or timeouts trigger NOTHING in Wombat!**

**Changes in the node family triggers adding or removing AntiDote nodes to / from
the cluster.**

**List global configurations**

1. The user requests the list of the global configurations. The server queries
the global_configuration bucket and returns its content.

**Change a global config**

1. The user requests changing a value of a global configuration.
Non-existing global configurations cannot be changed.
2. The process PUT
{set_config, Node, Application, [{Key, Val}], IsPersistent, new} to the
command bucket for each node. That means if a node family consists of 5 nodes,
then the process will PUT 5 new commands to the command bucket.
3. The plugin processes periodically GET the command bucket, therefore they will
indirectly be commanded to change a global config. If so, the plugin changes the
value of the given configuration (and persists it). After that it PUT its envs
to its local_configuration bucket. That will trigger further actions.
4. Finally, the plugin updates its command object and PUT the updated object to
the command bucket. That will trigger additional notifications pushed towards
Wombat.
5. Wombat waits for all the plugins to execute the command. Then it reports the
result to the user. Note that the result will be sent by a post-commit hook!
Once the changes have been performed on all the nodes, the post commit hook
of the global_configurations bucket notifies Wombat about the success (i.e. no
anomaly is related to the changed config).
6. The success is reported to the user.

**Reload the application**

1. The user requests reloading the application on all the nodes.
2. The process PUT {reload_application, Node,  Application, new} to the command
bucket.
3. The plugin processes periodically GET the command bucket, therefore they will
indirectly be commanded to reload the application.
5. Finally, the plugin updates the command object with the result and puts the
new object to the command bucket. Wombat will be notified.
6. The result is reported to the user.

**Examining the approach**

- Relies on a decentralised, distributed, replicated data store.
- During init the managed nodes must be reachable by the Wombats.
- During operation only one AntiDote node needs to be reachable by any of the
Wombats.
- Can handle network partitions much better.
