#!/bin/bash

cd "$(dirname "${BASH_SOURCE[0]}")"

readonly DEBUG_HOME_DIR="debug"
readonly WOMBAT_LOGS_DIR="_build/default/rel/wombat/log"
readonly WOMBAT_DATA_DIR="_build/default/rel/wombat/data"
readonly WOMBAT_CONFIG_FILES_DIR="_build/default/rel/wombat/releases"
readonly WOMBAT_REL_DIR="_build/default/rel"

current_debug_dir=""
fullpath_to_debug_dir=""

is_script_called_only_with_help_arg() {
    total_args=$1
    first_arg=$2
    if [[ $total_args -eq 1 && ($first_arg == "-h" || $first_arg == "--help") ]]; then
        echo 0 #true
    else
        echo 1 #false
    fi
}

print_help() {
    echo "Usage"
    echo "./wombat-debug.sh [-l] [-d] [-c] [-s] [-r]"
    echo "or"
    echo "./wombat-debug.sh [--logs] [--data] [--config] [--syscmd] [--root]"
}

create_target_directories() {
    echo "Creating target directories ../$DEBUG_HOME_DIR ...."
    mkdir -p "../$DEBUG_HOME_DIR"
    current_date_and_time=$(date +'%Y-%m-%d_%H-%M-%S')
    current_debug_dir="debug_$current_date_and_time"
    fullpath_to_debug_dir="../$DEBUG_HOME_DIR/$current_debug_dir"
    mkdir -p "$fullpath_to_debug_dir"
}

copy_directory() {
    echo "Copying $1 directory...."
    source_dir=$1
    fullpath_to_debug_dir=$2
    cp -r "../$source_dir" "$fullpath_to_debug_dir"
}

write_disk_usage_info() {
    echo "Writing disk usage output for $1"
    source_dir=$1
    fullpath_to_debug_dir=$2
    data_dir="$fullpath_to_debug_dir/data"
    mkdir -p $data_dir
    touch $data_dir/data_file_sizes
    find ../$source_dir -type f | xargs du -sh >> $data_dir/data_file_sizes
}

write_sys_command_output() {
    echo "Executing system commands...."
    fullpath_to_debug_dir=$1
    system_commands_debug_dir="$fullpath_to_debug_dir/system_commands"
    mkdir -p $system_commands_debug_dir
    write_output $system_commands_debug_dir/uname_a uname -a
    write_output $system_commands_debug_dir/ifconfig ifconfig
    write_output $system_commands_debug_dir/hostname hostname
    write_output $system_commands_debug_dir/w w
    write_output $system_commands_debug_dir/date date
    write_output $system_commands_debug_dir/df df
    write_output $system_commands_debug_dir/df_i df -i
    write_output $system_commands_debug_dir/dmesg dmesg
    write_output $system_commands_debug_dir/free_h free -h
    write_output $system_commands_debug_dir/iostat_mx iostat -mx
    write_output $system_commands_debug_dir/last last
    write_output $system_commands_debug_dir/mount mount
    write_output $system_commands_debug_dir/netstat_an netstat -an
    write_output $system_commands_debug_dir/netstat_i netstat -i
    write_output $system_commands_debug_dir/netstat_rn netstat -rn
    write_output $system_commands_debug_dir/ps_aux ps aux
    write_output $system_commands_debug_dir/sysctl_a sysctl -a
    [ `uname -s` == "Linux" ] && write_output $system_commands_debug_dir/lsb_release_a lsb_release -a
}

write_root_sys_command_output() {
    if [ `id -u` != 0 ]; then
        echo "Root system commands should be executed only as root! You're not root currently."
        return
    fi
    echo "Executing system commands as root...."
    fullpath_to_debug_dir=$1
    system_commands_debug_dir="$fullpath_to_debug_dir/system_commands"
    mkdir -p $system_commands_debug_dir
    [ `uname -s` == "Linux" ] && write_output $system_commands_debug_dir/iptables_L_nat iptables -L -t nat
    [ `uname -s` == "Linux" ] && write_output $system_commands_debug_dir/iptables_L iptables -L
}

write_output() {
    target_file=$1
    shift
    echo "command:$@" > "$target_file"
    echo "" >> "$target_file"
    "$@" >> "$target_file" 2>&1
    echo "" >> "$target_file"
    return_value=$?
    echo "return_value:$return_value" >> "$target_file"
}

compress_debug_directory() {
    curr_debug_dir=$1
    cd "../$DEBUG_HOME_DIR"
    tar_file_name="$curr_debug_dir.tar.gz"
    tar -zcf $tar_file_name "$curr_debug_dir"
    rm -rf "$curr_debug_dir"
    working_dir=$(pwd)
    cd ../scripts
    echo "$working_dir/$tar_file_name"
}

total_args=$#
if [[ $total_args -eq 0 ]]; then
    echo "Full path to debug dir $fullpath_to_debug_dir"
    create_target_directories
    copy_directory $WOMBAT_LOGS_DIR $fullpath_to_debug_dir
    write_disk_usage_info $WOMBAT_DATA_DIR $fullpath_to_debug_dir
    copy_directory $WOMBAT_CONFIG_FILES_DIR $fullpath_to_debug_dir
    write_sys_command_output $fullpath_to_debug_dir    
    [ `id -u` == 0 ] && write_root_sys_command_output $fullpath_to_debug_dir    
elif [[ $(is_script_called_only_with_help_arg $total_args $1) -eq 0 ]]; then
    print_help
else
    create_target_directories
    while [ -n "$1" ]; do
        case "$1" in
            -h|--help)
                print_help;;
            -l|--logs)
                copy_directory $WOMBAT_LOGS_DIR $fullpath_to_debug_dir;;
            -d|--data)
                write_disk_usage_info $WOMBAT_DATA_DIR $fullpath_to_debug_dir;;
            -c|--config)
                copy_directory $WOMBAT_CONFIG_FILES_DIR $fullpath_to_debug_dir;;
            -s|--syscmd)
                write_sys_command_output $fullpath_to_debug_dir;;
            -r|--root)
                write_root_sys_command_output $fullpath_to_debug_dir;;
            *)
                echo "Unknown option $1 encountered.";;
        esac
        shift
    done
fi

if [[ -d $fullpath_to_debug_dir ]]; then
    if [[ "$(ls -A $fullpath_to_debug_dir)" ]]; then
        tarball_location=$(compress_debug_directory $current_debug_dir)
        echo "Debug tarball created: $tarball_location"
    else
        # To remove the empty directory created when the script was
        # called with only invalid arguments.
        rm -rf "$fullpath_to_debug_dir"
    fi
fi
